@@ -17,6 +17,7 @@
 #include <zorp/log.h>
 #include <zorp/thread.h>
 #include <zorp/error.h>
+#include <zorp/misc.h>
 
 #include <stdarg.h>
 #include <stdio.h>
@@ -28,9 +29,6 @@
 #include <ctype.h>
 #include <stdlib.h>
 #include <sys/types.h>
-#ifdef HAVE_SYS_TIME_H
-  #include <sys/time.h>
-#endif
 #include <time.h>
 
 #ifdef G_OS_WIN32
@@ -41,14 +39,34 @@
 #  define close _close
 #  define open _open
 #else
+#  include <signal.h>
 #  include <sys/socket.h>
 #  include <sys/un.h>
 #endif
 
+#ifdef G_OS_WIN32  
+//Windows exception handler needs this 
+LONG WINAPI z_unhandled_exception_filter(PEXCEPTION_POINTERS);
+typedef LONG WINAPI z_unhandled_exception_filter_function(PEXCEPTION_POINTERS);
+z_unhandled_exception_filter_function *z_setup_unhandled_exception_filter = &z_unhandled_exception_filter;
+#endif
+
+
 #undef G_LOG_DOMAIN
 #define G_LOG_DOMAIN "Zorp"
 
-/* per-thread GHashTable based logtag cache */
+/**
+ * Extremal dummy value for booleans for detecting whether they are changed.
+ * Rationale: FALSE is 0, TRUE is some unknown value, none of 255, 256, 65535,
+ * 65536, (1L<<32)-1, (1L<<32) is divisible by 7, so this formula will yield
+ * a value different from both 0 and TRUE, even considering a byte, word or
+ * doubleword wrap-around :)
+ **/
+#define Z_EXTREMAL_BOOLEAN ((gboolean)(7*(int)TRUE + 1))
+
+/**
+ * Per-thread GHashTable based logtag cache.
+ **/
 typedef struct _ZLogTagCache
 {
   gboolean empty_hash;
@@ -56,26 +74,45 @@ typedef struct _ZLogTagCache
   GHashTable *tag_hash;
 } ZLogTagCache;
 
-/* parsed item of a ZLogSpec, pattern and verbosity_level */
+/**
+ * Parsed item of a ZLogSpec, pattern and verbosity_level.
+ **/
 typedef struct _ZLogSpecItem
 {
   gchar *pattern;
   gint verbose_level;
 } ZLogSpecItem;
 
-/* a complete logspec with a list of ZLogSpecItems */
+/**
+ * Complete logspec with a list of ZLogSpecItems.
+ **/
 typedef struct _ZLogSpec
 {
   GSList *items;
   gint verbose_level;
 } ZLogSpec;
 
-/* protects logtag_caches array when grabbing and releasing a thread specific log cache */
+/**
+ * Logging options.
+ **/
+typedef struct _ZLogOpts
+{
+  gint verbose_level;           /**< verbosity level */
+  gboolean use_syslog;          /**< whether to use syslog */
+  gboolean log_tags;            /**< whether to include message tag and verbosity level in log messages */
+  const gchar *log_spec;        /**< logspec specification */
+} ZLogOpts;
+
+ZLogOpts log_opts = {0, FALSE, FALSE, NULL};
+
+ZLogOpts log_opts_cmdline = {-1, Z_EXTREMAL_BOOLEAN, Z_EXTREMAL_BOOLEAN, NULL};
+
+/** Protects logtag_caches array when grabbing and releasing a thread specific log cache */
 static GStaticMutex logtag_cache_lock = G_STATIC_MUTEX_INIT;
 static GStaticPrivate current_logtag_cache = G_STATIC_PRIVATE_INIT;
 static GPtrArray *logtag_caches;
 
-/* protects log_spec/log_spec_str */
+/** Protects log_spec/log_spec_str */
 static GStaticMutex log_spec_lock = G_STATIC_MUTEX_INIT;
 static ZLogSpec log_spec;
 static gchar *log_spec_str;
@@ -85,12 +122,10 @@ static ZLogMapTagFunc log_map_tag;
 static gint log_mapped_tags_count;
 static guchar *log_mapped_tags_verb;
 
-static gboolean log_tags;
 static gboolean stderr_syslog = FALSE;
+static gboolean log_tags = FALSE;
 
-
-
-gchar fake_session_id[256] = "noname/nosession";
+gchar fake_session_id[256] = "nosession";
 
 static GMainContext *log_context = NULL;
 
@@ -109,12 +144,11 @@ int syslog_fd = -1;
 
 
 /**
- * z_open_syslog:
- * @tag: program name used in log messages
- *
  * Analogous to openlog(), open the syslog() connection to local syslogd. 
  *
- * Returns: TRUE when the operation succeeds
+ * @param[in] tag program name used in log messages
+ *
+ * @returns TRUE when the operation succeeds
  **/
 gboolean
 z_open_syslog(const gchar *tag)
@@ -122,7 +156,7 @@ z_open_syslog(const gchar *tag)
   struct sockaddr_un s_un;
   
   syslog_tag = tag;
-  syslog_fd = socket(AF_UNIX, SOCK_STREAM, 0);
+  syslog_fd = socket(PF_UNIX, SOCK_STREAM, 0);
   
   if (syslog_fd == -1)
     {
@@ -134,7 +168,7 @@ z_open_syslog(const gchar *tag)
   if (connect(syslog_fd, (struct sockaddr *) &s_un, sizeof(s_un)) == -1)
     {
       close(syslog_fd);
-      syslog_fd = socket(AF_UNIX, SOCK_DGRAM, 0);
+      syslog_fd = socket(PF_UNIX, SOCK_DGRAM, 0);
       if (connect(syslog_fd, (struct sockaddr *) &s_un, sizeof(s_un)) == -1)
         {
           close(syslog_fd);
@@ -146,14 +180,14 @@ z_open_syslog(const gchar *tag)
 }
 
 /**
- * z_close_syslog_internal:
- * @fd: syslog connection to close
+ * Close the connection to the local syslogd.
  *
- * Close the connection to the local syslogd. The syslog connection is
- * specified by the @fd argument to avoid races when reopening
+ * @param[in] fd syslog connection to close
+ *
+ * The syslog connection is specified by the fd argument to avoid races when reopening
  * connections.
  * 
- * Returns: whether the operation was succesful
+ * @returns whether the operation was successful
  **/
 gboolean
 z_close_syslog_internal(int fd)
@@ -167,11 +201,9 @@ z_close_syslog_internal(int fd)
 }
 
 /**
- * z_close_syslog:
- *
  * Close the connection to the global syslogd.
  *
- * Returns TRUE to indicate success.
+ * @returns TRUE to indicate success.
  **/
 gboolean
 z_close_syslog(void)
@@ -180,11 +212,10 @@ z_close_syslog(void)
 }
 
 /**
- * z_send_syslog:
- * @pri: syslog priority 
- * @msg: syslog message
- *
  * Send the specified message to syslog.
+ *
+ * @param[in] pri syslog priority 
+ * @param[in] msg syslog message
  **/
 gboolean
 z_send_syslog(gint pri, const gchar *msg)
@@ -198,7 +229,8 @@ z_send_syslog(gint pri, const gchar *msg)
   gchar timestamp[32];
   time_t now;
   struct tm t;
-  gint len, rc = 0, attempt = 0;
+  guint len, attempt = 0;
+  gint rc = 0;
   int sfd = syslog_fd;
   static GStaticMutex lock = G_STATIC_MUTEX_INIT;
   
@@ -211,7 +243,7 @@ z_send_syslog(gint pri, const gchar *msg)
   if (log_escape_nonprintable_chars)
     {
       len = strlen(buf);
-      for (p = msg; *p && len < sizeof(buf) - 5; p++)
+      for (p = (guchar *) msg; *p && len < sizeof(buf) - 5; p++)
         {
           if (*p >= 0x20 && *p <= 0x7F)
             {
@@ -258,11 +290,13 @@ z_send_syslog(gint pri, const gchar *msg)
 /* use the syslog implementation in libc */
 
 /**
- * z_open_syslog:
- * @tag: program name to display in log messages
+ * Open a connection to the local system logging process.
  *
- * Open a connection to the local system logging process. This function uses
- * the native syslog() implementation in libc to offer syslog functions.
+ * @param[in] tag program name to display in log messages
+ *
+ * This function uses the native syslog() implementation in libc to offer syslog functions.
+ *
+ * @returns always TRUE.
  **/
 gboolean
 z_open_syslog(const gchar *tag)
@@ -272,9 +306,9 @@ z_open_syslog(const gchar *tag)
 }
 
 /**
- * z_close_syslog:
- *
  * Close the connection to the local syslog process. 
+ *
+ * @returns always TRUE.
  **/
 gboolean
 z_close_syslog(void)
@@ -284,11 +318,12 @@ z_close_syslog(void)
 }
 
 /**
- * z_send_syslog:
- * @pri: priority of the message 
- * @msg: message itself
- * 
  * Send a message to syslogd. It uses the syslog() function in libc.
+ *
+ * @param[in] pri priority of the message 
+ * @param[in] msg message itself
+ *
+ * @returns always TRUE.
  **/
 gboolean
 z_send_syslog(gint pri, const gchar *msg)
@@ -304,13 +339,13 @@ z_send_syslog(gint pri, const gchar *msg)
 int syslog_fd = -1;
 
 /**
- * z_open_syslog:
- * @tag: program name used in log messages
- *
  * Analogous to openlog(), open the syslog() connection to local syslogd. 
+ *
+ * @param[in] tag program name used in log messages
+ *
  * Windows implementation.
  *
- * Returns: TRUE when the operation succeeds
+ * @returns TRUE when the operation succeeds
  **/
 gboolean
 z_open_syslog(const gchar *tag)
@@ -331,8 +366,6 @@ z_open_syslog(const gchar *tag)
 }
 
 /**
- * z_close_syslog:
- *
  * Close the connection to the local syslog process. 
  **/
 gboolean
@@ -351,11 +384,10 @@ z_close_syslog(void)
 /* logspec parsing and evaluation handling */
 
 /**
- * z_log_glob_match:
- * @glob: pattern to match @tag against
- * @tag: message tag
- *
  * Check when the pattern in a logspec matches the specified tag.
+ *
+ * @param[in] glob pattern to match tag against
+ * @param[in] tag message tag
  **/
 static gboolean
 z_log_spec_glob_match(const gchar *glob, const gchar *tag)
@@ -399,12 +431,13 @@ z_log_spec_glob_match(const gchar *glob, const gchar *tag)
 }
 
 /**
- * z_log_spec_eval:
- * @self: ZLogSpec structure
- * @tag: message to return verbosity for
+ * Evaluate the currently parsed logspec in self and return the verbosity level
+ * associated with tag.
  *
- * Evaluate the currently parsed logspec in @self and return the verbosity level
- * associated with @tag.
+ * @param[in] self ZLogSpec structure
+ * @param[in] tag message to return verbosity for
+ *
+ * @returns the verbosity level associated with tag
  **/
 static gint
 z_log_spec_eval(ZLogSpec *self, const gchar *tag)
@@ -426,11 +459,11 @@ z_log_spec_eval(ZLogSpec *self, const gchar *tag)
 }
 
 /**
- * z_log_spec_destroy:
- * @self: ZLogSpec instance
+ * Free all resources associated with self.
  *
- * Free all resources associated with @self. Does not free @self as it is assummed
- * to be allocated statically.
+ * @param[in] self ZLogSpec instance
+ *
+ * Does not free self as it is assumed to be allocated statically.
  **/
 static void
 z_log_spec_destroy(ZLogSpec *self)
@@ -453,12 +486,13 @@ z_log_spec_destroy(ZLogSpec *self)
 }
 
 /**
- * z_log_spec_init:
- * @self: ZLogSpec instance
- * @logspec_str: logspec specification
- * @default_verbosity: global verbosity level
+ * Parse a user-specified logspec into self.
  *
- * Parse a user-specified logspec into @self.
+ * @param[in] self ZLogSpec instance
+ * @param[in] logspec_str logspec specification
+ * @param[in] default_verbosity global verbosity level
+ *
+ * @returns sdf
  **/
 static gboolean
 z_log_spec_init(ZLogSpec *self, const gchar *logspec_str, gint default_verbosity)
@@ -522,11 +556,8 @@ z_log_spec_init(ZLogSpec *self, const gchar *logspec_str, gint default_verbosity
  * */
 
 /**
- * z_log_clear_caches:
- *
  * Clear all thread specific caches. It is called after changing the
  * verbosity level or the logspec.
- * 
  **/
 void
 z_log_clear_caches(void)
@@ -548,8 +579,6 @@ z_log_clear_caches(void)
 }
 
 /**
- * z_log_grab_cache:
- *
  * Grab a thread specific log-tag cache.
  **/
 void
@@ -583,8 +612,6 @@ z_log_grab_cache(void)
 }
 
 /**
- * z_log_grab_cache:
- *
  * Release a thread specific log-tag cache to make it usable by other threads.
  **/
 void
@@ -602,11 +629,10 @@ z_log_release_cache(void)
 }
 
 /**
- * z_log_thread_started:
- * @thread:
- * @user_data:
- *
  * Thread startup function called at thread startup time to allocate a logtag cache.
+ *
+ * @param thread (unused)
+ * @param user_data (unused)
  **/
 static void
 z_log_thread_started(ZThread *thread G_GNUC_UNUSED, gpointer user_data G_GNUC_UNUSED)
@@ -615,11 +641,10 @@ z_log_thread_started(ZThread *thread G_GNUC_UNUSED, gpointer user_data G_GNUC_UN
 }
 
 /**
- * z_log_thread_started:
- * @thread:
- * @user_data:
- *
  * Thread startup function called at thread shutdown time to release the logtag cache.
+ *
+ * @param thread (unused)
+ * @param user_data (unused)
  **/
 static void
 z_log_thread_stopped(ZThread *thread G_GNUC_UNUSED, gpointer user_data G_GNUC_UNUSED)
@@ -630,12 +655,14 @@ z_log_thread_stopped(ZThread *thread G_GNUC_UNUSED, gpointer user_data G_GNUC_UN
 /* global log state manipulation */
 
 /**
- * z_log_change_verbose_level:
- * @direction: specifies the change direction (-1 decrease, 0 set, 1 increase)
- * @value: change with this value
- *
  * This function changes the global verbosity level as specified by the
- * @direction and @value parameters.
+ * direction and value parameters.
+ *
+ * @param[in]  direction specifies the change direction (-1 decrease, 0 set, 1 increase)
+ * @param[in]  value change with this value (or set to this value if direction == 0)
+ * @param[out] new_value if not NULL, the resulting verbosity level is returned here
+ *
+ * @returns always TRUE.
  **/
 gboolean
 z_log_change_verbose_level(gint direction, gint value, gint *new_value)
@@ -663,15 +690,18 @@ z_log_change_verbose_level(gint direction, gint value, gint *new_value)
        */
       z_log(NULL, CORE_INFO, 0, "Changing verbosity level; verbose_level='%d'", log_spec.verbose_level);
     }
-  *new_value = log_spec.verbose_level;
+  if (new_value)
+    *new_value = log_spec.verbose_level;
   return TRUE;
 }
 
 /**
- * z_log_change_logspec:
- * @newspec: change the logspec to this value, leave it unchanged if set to NULL
- *
  * Change the logspec value and return the new setting.
+ *
+ * @param[in]  new_log_spec_str change the logspec to this value, leave it unchanged if set to NULL
+ * @param[out] new_value if not NULL, the new logspec will be returned here
+ *
+ * @return TRUE on success
  **/
 gboolean
 z_log_change_logspec(const gchar *new_log_spec_str, const gchar **new_value)
@@ -704,18 +734,19 @@ z_log_change_logspec(const gchar *new_log_spec_str, const gchar **new_value)
         }
     }
     
-  *new_value = log_spec_str;
+  if (new_value)
+    *new_value = log_spec_str;
   return TRUE;
 }
 
 /**
- * z_log_enable_tag_map_cache:
- * @map_tags: function to map message tags to IDs
- * @max_tag: maximum ID value assigned to tags
- *
  * This function enables the "tag_map cache" which makes tag caching very
  * efficient by using an array based lookup instead of GHashTable. 
- * NOTE: this function can only be called once, at startup time.
+ *
+ * @param[in] map_tags function to map message tags to IDs
+ * @param[in] max_tag maximum ID value assigned to tags
+ *
+ * @note this function can only be called once, at startup time.
  **/
 void
 z_log_enable_tag_map_cache(ZLogMapTagFunc map_tags, gint max_tag)
@@ -728,21 +759,22 @@ z_log_enable_tag_map_cache(ZLogMapTagFunc map_tags, gint max_tag)
 }
 
 /**
- * z_log_enabled_len:
- * @tag: log message tag
- * @tag_len: length of @tag
- * @level: log message level
- *
  * Checks if a message with a given class/level combination would actually
- * be written to the log. It can be used prior to constructing complex log
+ * be written to the log.
+ *
+ * @param[in] tag log message tag
+ * @param[in] tag_len length of tag
+ * @param[in] level log message level
+ *
+ * It can be used prior to constructing complex log
  * messages to decide whether the messages need to be constucted at all.
  * All results are cached, thus the second invocation will not parse the 
  * log specifications again.
  *
- * Returns: TRUE if the log would be written, FALSE otherwise
+ * @returns TRUE if the log would be written, FALSE otherwise
  **/
 gboolean
-z_log_enabled_len(gchar *tag, gsize tag_len, int level)
+z_log_enabled_len(const gchar *tag, gsize tag_len, gint level)
 {
   gint verbose;
   ZLogTagCache *lc;
@@ -767,7 +799,7 @@ z_log_enabled_len(gchar *tag, gsize tag_len, int level)
             {
               g_static_mutex_lock(&log_spec_lock);
               verbose = z_log_spec_eval(&log_spec, tag);
-              log_mapped_tags_verb[tag_ndx] = verbose + 1;
+              log_mapped_tags_verb[tag_ndx] = (guchar) (verbose & 0xFF) + 1;
               g_static_mutex_unlock(&log_spec_lock);
             }
           return level <= verbose;
@@ -786,14 +818,14 @@ z_log_enabled_len(gchar *tag, gsize tag_len, int level)
       lc->empty_hash = FALSE;
     }
   tag_hash = lc->tag_hash;
-  verbose = GPOINTER_TO_UINT(g_hash_table_lookup(tag_hash, tag));
+  verbose = GPOINTER_TO_INT(g_hash_table_lookup(tag_hash, (gconstpointer) tag));
   if (!verbose)
     {
       /* slooooww path, evaluate logspec */
       g_static_mutex_lock(&log_spec_lock);
       verbose = z_log_spec_eval(&log_spec, tag);
       g_static_mutex_unlock(&log_spec_lock);
-      g_hash_table_insert(tag_hash, tag, GUINT_TO_POINTER(verbose + 1));
+      g_hash_table_insert(tag_hash, (gchar *) tag, GUINT_TO_POINTER(verbose + 1));
     }
   else
     verbose--;
@@ -804,12 +836,15 @@ z_log_enabled_len(gchar *tag, gsize tag_len, int level)
 /* Main entry points for logging */
 
 /**
- * z_log_session_id:
- * @session_id: default session_id
+ * Get the current session id.
  *
- * This helper function is used by the z_log() macro to get the current
- * session id. If the argument is NULL the session_id assigned to the
- * current thread is used, otherwise the value of the argument is returned.
+ * @param[in] session_id default session_id
+ *
+ * This helper function is used by the z_log() macro (or function, in
+ * the case of Win32) to get the current session id.
+ *
+ * @returns If the argument is NULL, the session_id assigned to the
+ * current thread, otherwise the value of the argument is returned.
  **/
 const gchar *
 z_log_session_id(const gchar *session_id)
@@ -827,26 +862,27 @@ z_log_session_id(const gchar *session_id)
 
 
 /**
- * z_logv2:
- * @class: log message class
- * @level: log message verbosity level
- * @format: log message format specified in printf form
- * @ap: format arguments va_list
- *
  * This function sends a message formatted as printf format string and
- * arguments to the syslog. The associated class/level pair is checked
+ * arguments to the syslog.
+ *
+ * @param[in] class log message class
+ * @param[in] level log message verbosity level
+ * @param[in] format log message format specified in printf form
+ * @param[in] ap format arguments va_list
+ *
+ * The associated class/level pair is checked
  * whether the message really needs to be written.
  **/
 void
-z_logv2(gchar *class, int level, gchar *format, va_list ap)
+z_logv(const gchar *class, int level, gchar *format, va_list ap)
 {
   int saved_errno = errno;
   
   if (log_tags)
     {
       /*
-       * g_log cut it in 1024. byte,
-       * so eonught to write this amount of data.
+       * g_log cut it in the 1024th byte,
+       * so it's enough to write this amount of data.
        */
       gchar msgbuf[1024];
       g_vsnprintf(msgbuf, sizeof(msgbuf), format, ap);
@@ -865,35 +901,17 @@ z_logv2(gchar *class, int level, gchar *format, va_list ap)
 }
 
 /**
- * z_logv:
- * @class: log message class
- * @level: log message verbosity level
- * @format: log message format specified in printf form
- * @ap: format arguments va_list
- *
- * This function sends a message formatted as printf format string and
- * arguments to the syslog. The associated class/level pair is checked
- * whether the message really needs to be written.
- **/
-void
-z_logv(gchar *class, int level, gchar *format, va_list ap)
-{
-  if (z_log_enabled(class, level))
-    z_logv2(class, level, format, ap);
-}
-
-
-/**
- * z_llog:
- * @class: log message class
- * @level: log message verbosity level
- * @format: log message format specified in printf form
- *
  * This message is the same as z_logv() but format string and arguments
  * are specified directly.
+ *
+ * @param[in] class log message class
+ * @param[in] level log message verbosity level
+ * @param[in] format log message format specified in printf form
+ *
+ * @see z_log() in log.h
  **/
 void
-z_llog(gchar *class, int level, gchar *format, ...)
+z_llog(const gchar *class, int level, gchar *format, ...)
 {
   va_list l;
 
@@ -902,34 +920,28 @@ z_llog(gchar *class, int level, gchar *format, ...)
   va_end(l);
 }
 
-/**
- * z_llog2:
- * @class: log message class
- * @level: log message verbosity level
- * @format: log message format specified in printf form
- *
- * This message is the same as z_logv() but format string and arguments
- * are specified directly.
- **/
-void
-z_llog2(gchar *class, int level, gchar *format, ...)
-{
-  va_list l;
-
-  va_start(l, format);
-  z_logv2(class, level, format, l);
-  va_end(l);
-}
-
 #ifdef G_OS_WIN32
 
+/**
+ * Win32 implementation of z_log().
+ *
+ * @param[in] session_id session id
+ * @param[in] class log message class
+ * @param[in] level log message verbosity level
+ * @param[in] format log message format specified in printf form
+ *
+ * This function checks if the message would be logged; if so, sends
+ * a message formatted as printf format string and
+ * arguments (including session id) to the syslog.
+ *
+ * @see z_log() in log.h
+ **/
 void 
-z_log(gchar *session_id, gchar *class, int level, gchar *format, ...)
+z_log(const gchar *session_id, const gchar *class, int level, gchar *format, ...)
 {
   va_list l;
   gchar msgbuf[2048];
 
-
   if (!z_log_enabled(class, level))
     return;
 
@@ -937,23 +949,50 @@ z_log(gchar *session_id, gchar *class, int level, gchar *format, ...)
 
   g_vsnprintf(msgbuf, sizeof(msgbuf), format, l);
 
-  g_log(G_LOG_DOMAIN, G_LOG_LEVEL_INFO, "%p -> %s(%d): %s", g_thread_self(), class, level, msgbuf);
+  g_log(G_LOG_DOMAIN, G_LOG_LEVEL_INFO, "%p -> %s(%d): (%s): %s", g_thread_self(),
+        class, level, z_log_session_id(session_id), msgbuf);
   va_end(l);
 }
 
 #endif /* G_OS_WIN32 */
 
+/**
+ * This function is registered as a GLib log handler and sends all GLIB
+ * messages to stderr.
+ *
+ * @param     log_domain not used
+ * @param     log_flags not used
+ * @param[in] message message
+ * @param     user_data not used 
+ **/
+static void
+z_log_func_nosyslog(const gchar *log_domain G_GNUC_UNUSED,
+	   GLogLevelFlags log_flags G_GNUC_UNUSED,
+	   const gchar *message,
+	   gpointer user_data G_GNUC_UNUSED)
+{
+  gchar timestamp[32];
+  time_t now;
+  struct tm tmnow;
+
+  /* prepend timestamp */
+  time(&now);
+  strftime(timestamp, sizeof(timestamp), "%b %d %H:%M:%S", localtime_r(&now, &tmnow));
+  fprintf(stderr, "%s %s\n", timestamp, message);
+}
+
+
 #ifndef G_OS_WIN32
 /**
- * z_log_func:
- * @log_domain: GLIB log domain
- * @log_flags: GLIB log flags
- * @message: message
- * @user_data: not used 
- *
  * This function is registered as a GLib log handler and sends all GLIB
- * messages to messages to syslog. Zorp itself does not use GLIB logging
- * it calls z_log() directly.
+ * messages to messages to syslog.
+ *
+ * @param     log_domain GLIB log domain (unused)
+ * @param[in] log_flags GLIB log flags
+ * @param[in] message message
+ * @param     user_data not used 
+ *
+ * Zorp itself does not use GLIB logging, it calls z_log() directly.
  **/
 static void
 z_log_func(const gchar *log_domain G_GNUC_UNUSED,
@@ -974,40 +1013,77 @@ z_log_func(const gchar *log_domain G_GNUC_UNUSED,
 
 #else
 
-/*+ log handler function to send Win32 debug message +*/
+static GStaticMutex win32_log_handler_mutex = G_STATIC_MUTEX_INIT;
+
+/**
+ * Log handler function to send Win32 debug message.
+ *
+ * @param     log_domain unused
+ * @param     log_flags unused
+ * @param[in] message debug message to send
+ * @param     user_data unused
+ **/
 static void
 z_log_win32_debugmsg(const gchar *log_domain,
                  GLogLevelFlags  log_flags,
                     const gchar *message,
                        gpointer  user_data)
 {
+  g_static_mutex_lock(&win32_log_handler_mutex);
+
   OutputDebugString(message);
   OutputDebugString("\n");
+
+  g_static_mutex_unlock(&win32_log_handler_mutex);
 }
 
+/**
+ * Log handler function to send Win32 syslog message.
+ *
+ * @param     log_domain unused
+ * @param     log_flags unused
+ * @param[in] message syslog message to send
+ * @param     user_data unused
+ **/
 static void
 z_log_win32_syslogmsg(const gchar *log_domain,
                  GLogLevelFlags  log_flags,
                     const gchar *message,
                        gpointer  user_data)
 {
-  write(syslog_fd, message, strlen(message));
-  write(syslog_fd, "\n", 1);
+  time_t now;
+  struct tm *t;
+  gchar tstamp[64];
+  gchar buf[2048];
+  int nchars;
+ 
+  g_static_mutex_lock(&win32_log_handler_mutex);
+
+  now = time(NULL);
+  t = localtime(&now);
+  strftime(tstamp, sizeof(tstamp), "%Y %b %d %H:%M:%S", t);
+
+  g_static_mutex_unlock(&win32_log_handler_mutex);
+
+  nchars = g_snprintf(buf, sizeof(buf), "%s %s\n", tstamp, message);
+
+  write(syslog_fd, buf, nchars);
 }
 
 #endif
 
 /**
- * z_fetch_stderr:
- * @channel: the read end of the STDERR pipe
- * @condition: the I/O condition triggering this callback
- * @arg: not used
+ * Fetch messages line-by-line and send them to log via z_log().
+ * 
+ * @param[in] channel the read end of the STDERR pipe
+ * @param     condition the I/O condition triggering this callback (unused)
+ * @param     arg not used
  *
  * This function is registered as a read callback of the STDERR pipe to 
  * fetch messages sent to stderr. It fetches messages line-by-line and
  * uses z_log() to send messages to log.
  *
- * Returns: TRUE to indicate further reading is needed, FALSE otherwise
+ * @returns TRUE to indicate further reading is needed, FALSE otherwise
  **/
 gboolean
 z_fetch_stderr(GIOChannel *channel, GIOCondition condition G_GNUC_UNUSED, gpointer arg G_GNUC_UNUSED)
@@ -1045,12 +1121,11 @@ z_fetch_stderr(GIOChannel *channel, GIOCondition condition G_GNUC_UNUSED, gpoint
 }
 
 /**
- * z_log_source_new:
- * @fd: read side of the stderr pipe
- *
  * Creates the source watching the stderr pipe.
+ *
+ * @param[in] fd read side of the stderr pipe
  **/
-void
+GSource *
 z_log_source_new(gint fd)
 {
   GIOChannel *channel;
@@ -1061,28 +1136,40 @@ z_log_source_new(gint fd)
   g_io_channel_set_flags(channel, G_IO_FLAG_NONBLOCK, NULL);
   source = g_io_create_watch(channel, G_IO_IN);
   g_source_set_callback(source, (GSourceFunc) z_fetch_stderr, NULL, NULL);
-  g_source_attach(source, log_context);
-  g_source_unref(source);
+  return source;
 }
 
 /**
- * z_log_run:
- * @user_data: thread data pointer, assumed to point to the stderr fd
- *
  * STDERR reading thread function.
+ *
+ * @param[in] user_data thread data pointer, assumed to point to the stderr fd
+ *
+ * @returns always NULL.
  **/
 void *
 z_log_run(gpointer user_data)
 {
   GMainContext *c, *old_context;
   gint *fd = (gint *)user_data;
+  GSource *source;
+  
+  old_context = g_main_context_new();
+  g_main_context_ref(old_context);
+  log_context = old_context;
   
-  old_context = log_context = g_main_context_new();
   g_main_context_acquire(log_context);
-  z_log_source_new(*fd);
+  
+  source = z_log_source_new(*fd);
+  g_source_attach(source, log_context);
+  g_source_unref(source);
 
   do
     {
+      /* NOTE: although log_context might already be unrefed by the thread
+       * running z_log_destroy() we still hold a reference through
+       * old_context, therefore it is not a problem if we use a stale value
+       * in log_context, at most we run another iteration */
+      
       c = log_context;
   
       if (c)
@@ -1092,15 +1179,15 @@ z_log_run(gpointer user_data)
 
   g_main_context_release(old_context);
   g_main_context_unref(old_context);
+
   return NULL;
 }
 
 /**
- * z_log_enable_syslog:
- * @syslog_name: syslog program name
- * 
  * This function can be called after z_log_init() to redirect internal
  * messages to syslog.
+ *
+ * @param[in] syslog_name syslog program name
  **/
 void
 z_log_enable_syslog(const gchar *syslog_name)
@@ -1114,11 +1201,10 @@ z_log_enable_syslog(const gchar *syslog_name)
 }
 
 /**
- * z_log_enable_stderr_redirect:
- * @threaded: specifies whether to use a separate thread for stderr reading
- * 
  * This function can be called after z_log_init() to redirect stderr
  * messages messages to the system log.
+ *
+ * @param[in] threaded specifies whether to use a separate thread for stderr reading
  **/
 void
 z_log_enable_stderr_redirect(gboolean threaded)
@@ -1143,7 +1229,7 @@ z_log_enable_stderr_redirect(gboolean threaded)
 
   if (threaded)
     {
-      if (!z_thread_new("Log thread", z_log_run, &grab[0]))
+      if (!z_thread_new("stderr", z_log_run, &grab[0]))
         threaded = FALSE;
     }
 
@@ -1163,34 +1249,49 @@ z_log_enable_stderr_redirect(gboolean threaded)
 
 
 /**
- * z_log_init:
- * @ls: log specification
- * @verbose_level: global verbose level
- * @syslog_name: the program name to appear in syslogs
- * @flags: log flags (ZLF_* macros)
- *
  * Initialize the logging subsystem according to the options
- * in specified in the @flags parameter.
+ * in specified in the flags parameter.
+ *
+ * @param[in] syslog_name the program name to appear in syslogs
+ * @param[in] flags log flags (ZLF_* macros)
+ *
+ * @returns TRUE on success
  **/
 gboolean
-z_log_init(const gchar *ls, gint verbose_level, const gchar *syslog_name, guint flags)
+z_log_init(const gchar *syslog_name, guint flags)
 {
-  if (!z_log_spec_init(&log_spec, ls, verbose_level))
+#ifndef G_OS_WIN32
+  struct sigaction sa;
+  gint i;
+
+  i = sigaction(SIGPIPE, NULL, &sa);
+  if (i)
     {
-      z_log(NULL, CORE_ERROR, 0, "Invalid logspec; logspec='%s'", ls);
+      z_log(NULL, CORE_ERROR, 0, "Can't get SIGPIPE handler; error='%s'", strerror(errno));
+    }
+  else if (sa.sa_handler == SIG_DFL)
+    {
+      sa.sa_handler = SIG_IGN;
+      if (sigaction(SIGPIPE, &sa, NULL))
+        z_log(NULL, CORE_ERROR, 0, "Can't set SIGPIPE handler; error='%s'", strerror(errno));
+    }
+#endif
+
+  if (!z_log_spec_init(&log_spec, z_log_get_log_spec(), z_log_get_verbose_level()))
+    {
+      z_log(NULL, CORE_ERROR, 0, "Invalid logspec; logspec='%s'", z_log_get_log_spec());
       return FALSE;
     }
-  log_spec_str = ls ? g_strdup(ls) : NULL;
+  log_spec_str = z_log_get_log_spec() ? g_strdup(z_log_get_log_spec()) : NULL;
+  log_tags = z_log_get_log_tags();
     
-  log_tags = !!(flags & ZLF_TAGS);
-  
   logtag_caches = g_ptr_array_new();
   
   z_log_grab_cache();
   z_thread_register_start_callback((GFunc) z_log_thread_started, NULL);
   z_thread_register_stop_callback((GFunc) z_log_thread_stopped, NULL);
 
-  if (flags & ZLF_SYSLOG)
+  if (z_log_get_use_syslog())
     {
       z_log_enable_syslog(syslog_name);
 
@@ -1199,12 +1300,19 @@ z_log_init(const gchar *ls, gint verbose_level, const gchar *syslog_name, guint
         z_log_enable_stderr_redirect(flags & ZLF_THREAD);
 #endif
     }
-#ifdef G_OS_WIN32
-  else if (flags & ZLF_WINDEBUG)
+  else
     {
-      g_log_set_handler(G_LOG_DOMAIN, 0xff, z_log_win32_debugmsg, NULL);
-    }
+#ifdef G_OS_WIN32
+      if (flags & ZLF_WINDEBUG)
+        {
+          g_log_set_handler(G_LOG_DOMAIN, 0xff, z_log_win32_debugmsg, NULL);
+        }
+      else
 #endif  
+        {
+          g_log_set_handler(G_LOG_DOMAIN, 0xff, z_log_func_nosyslog, NULL);
+        }
+    }
 
   if (flags & ZLF_ESCAPE)
     log_escape_nonprintable_chars = TRUE;
@@ -1212,8 +1320,6 @@ z_log_init(const gchar *ls, gint verbose_level, const gchar *syslog_name, guint
 }
 
 /**
- * z_log_destroy:
- *
  * Deinitialize the logging subsystem.
  **/
 void
@@ -1229,7 +1335,162 @@ z_log_destroy(void)
 #endif
   z_close_syslog();
   
+  /* NOTE: log_context is freed in the log thread */
   c = log_context;
   log_context = NULL;
-  g_main_context_wakeup(c);
+  if (c)
+    {
+      g_main_context_wakeup(c);
+      g_main_context_unref(c);
+    }
+}
+
+/**
+ * Keep track of indentation level and return the appropriate number of spaces.
+ *
+ * @param[in] dir direction and amount to indent
+ *
+ * The indentation level has to be in [0, 128], changes beyond that won't be performed.
+ * The indentation level is tracked separately per thread.
+ *
+ * @returns pointer to static array of a number of spaces equal to the indentation level
+ **/
+const gchar *
+z_log_trace_indent(gint dir)
+{
+  static const gchar *spaces128 =
+    "                                                                "
+    "                                                                ";
+    
+  static GStaticPrivate current_indent_key = G_STATIC_PRIVATE_INIT;
+  int *current_indent = g_static_private_get (&current_indent_key);
+  const gchar *res;
+
+  if (!current_indent)
+    {
+      current_indent = g_new (int,1);
+      *current_indent = 0;
+      g_static_private_set (&current_indent_key, current_indent, g_free);
+    }
+
+  if (dir > 0)
+    {
+      res = spaces128 + 128 - *current_indent;
+      if (*current_indent < (128 - dir))
+        *current_indent += dir;
+    }
+  else
+    {
+      if (*current_indent >= -dir)
+        *current_indent += dir;
+      res = spaces128 + 128 - *current_indent;
+    }
+  return res;
+}
+
+/**
+ * Command line options for logging.
+ **/
+static GOptionEntry z_log_option_entries[] =
+{
+  { "verbose",    'v',                          0, G_OPTION_ARG_INT,      &log_opts_cmdline.verbose_level,    "Set verbosity level", "<verbosity>" },
+  { "no-syslog",  'l',      G_OPTION_FLAG_REVERSE, G_OPTION_ARG_NONE,     &log_opts_cmdline.use_syslog,       "Do not send messages to syslog", NULL },
+  { "log-spec",   's',                          0, G_OPTION_ARG_STRING,   &log_opts_cmdline.log_spec,         "Set log specification", "<logspec>" },
+  { "logspec",    's',      G_OPTION_FLAG_HIDDEN, G_OPTION_ARG_STRING,    &log_opts_cmdline.log_spec,         "Alias for log-spec", "<logspec>" },
+  { "log-tags",   'T',                          0, G_OPTION_ARG_NONE,     &log_opts_cmdline.log_tags,         "Enable logging of message tags", NULL },
+  { NULL,          0,                           0,                   0,   NULL, NULL, NULL }
+};
+
+/**
+ * Set default options in log_opts.
+ *
+ * @param[in] verbose_level verbosity level
+ * @param[in] use_syslog whether to use syslog
+ * @param[in] n_log_tags whether to include message tag and verbosity level in log messages
+ * @param[in] n_log_spec logspec specification
+ **/
+void
+z_log_set_defaults(gint verbose_level, gboolean use_syslog, gboolean n_log_tags, const gchar *n_log_spec)
+{
+  log_opts.verbose_level = verbose_level;
+  log_opts.use_syslog = use_syslog;
+  log_opts.log_tags = n_log_tags;
+  log_opts.log_spec = n_log_spec;
+}
+
+/**
+ * Add the logging-specific command line options to the option context.
+ *
+ * @param[in] ctx GOptionContext instance
+ **/
+void
+z_log_add_option_group(GOptionContext *ctx)
+{
+  GOptionGroup *group;
+  
+  /* initialise commandline arg variables to extremal values to be able to detect
+   * whether they are changed or not */
+  log_opts_cmdline.verbose_level = -1;
+  log_opts_cmdline.use_syslog = Z_EXTREMAL_BOOLEAN;
+  log_opts_cmdline.log_spec = NULL;
+  log_opts_cmdline.log_tags = Z_EXTREMAL_BOOLEAN;
+  
+  group = g_option_group_new("log", "Log options", "Log options", NULL, NULL);
+  g_option_group_add_entries(group, z_log_option_entries);
+  g_option_context_add_group(ctx, group);
+}
+
+/**
+ * Get verbose level setting in effect.
+ *
+ * @returns the setting from the command line option if one was given and the default otherwise.
+ **/
+gint
+z_log_get_verbose_level(void)
+{
+  return (log_opts_cmdline.verbose_level == -1) ? log_opts.verbose_level : log_opts_cmdline.verbose_level;
+}
+
+/**
+ * Get syslog setting in effect.
+ *
+ * @returns the setting from the command line option if one was given and the default otherwise.
+ **/
+gboolean
+z_log_get_use_syslog(void)
+{
+  return (log_opts_cmdline.use_syslog == Z_EXTREMAL_BOOLEAN) ? log_opts.use_syslog : log_opts_cmdline.use_syslog;
+}
+
+/**
+ * Get log tags setting in effect.
+ *
+ * @returns the setting from the command line option if one was given and the default otherwise.
+ **/
+const gchar *
+z_log_get_log_spec(void)
+{
+  return (log_opts_cmdline.log_spec == NULL) ? log_opts.log_spec : log_opts_cmdline.log_spec;
+}
+
+/**
+ * Get logspec string in effect.
+ *
+ * @returns the setting from the command line option if one was given and the default otherwise.
+ **/
+gboolean
+z_log_get_log_tags(void)
+{
+  return (log_opts_cmdline.log_tags == Z_EXTREMAL_BOOLEAN) ? log_opts.log_tags : log_opts_cmdline.log_tags;
+}
+
+/**
+ * Set the value of the use_syslog option.
+ *
+ * @param[in] use_syslog the new value
+ **/
+void
+z_log_set_use_syslog(gboolean use_syslog)
+{
+  log_opts.use_syslog = use_syslog;
 }
