@@ -14,6 +14,7 @@
  *
  ***************************************************************************/
 
+#include <zorp/misc.h>
 #include <zorp/zorplib.h>
 
 #if ZORPLIB_ENABLE_MEM_TRACE
@@ -33,54 +34,58 @@
 #endif
 #include <stdio.h>
 #include <time.h>
-#ifdef HAVE_SYS_TIME_H
-  #include <sys/time.h>
-#endif
 #include <string.h>
 
+#if HAVE_BACKTRACE
+#  include <execinfo.h>
+#else
+# define backtrace(a, b) {a[0]=NULL}
+#endif
+
 #define MEMTRACE_BACKTRACE_LEN 64
-#define MEMTRACE_BACKTRACE_BUF_LEN (MEMTRACE_BACKTRACE_LEN * 11 + 1)
+#define MEMTRACE_BACKTRACE_BUF_LEN (MEMTRACE_BACKTRACE_LEN * ((sizeof(gpointer) * 2) + 2 + 1) + 1)
 
 #define MEMTRACE_CANARY_SIZE 2
 #define MEMTRACE_CANARY_FILL 0xcd
 #define MEMTRACE_CANARY_CHECK 0xcdcdcdcd
 #define MEMTRACE_CANARY_OVERHEAD sizeof(ZMemTraceCanary) * 2
 
-/*
+/**
  * Trace the memory usage of the program using libzorpll
- * FIXMEE May we use valgind instead?
- */
+ *
+ * @todo FIXME: May we use valgrind instead?
+ **/
 #define ZORP_ENV_MEMTRACE "ZORP_MEMTRACE"
 
-/* 
+/**
  * Set some canaries before and after the allocated memory
  * and check it when a pointer deallocated.
- */
+ **/
 #define ZORP_ENV_MEMTRACE_CANARIES "ZORP_MEMTRACE_CANARIES"
 
-/*
+/**
  * Don't deallocate pointers, just fill it with the canaries
- * It's usefull when checking for heap corruption
- */
+ * It's useful when checking for heap corruption
+ **/
 #define ZORP_ENV_MEMTRACE_HARD     "ZORP_MEMTRACE_HARD"
 
-/*
- * Log all allocation and deallocation event.
- */
+/**
+ * Log all allocation and deallocation events.
+ **/
 #define ZORP_ENV_MEMTRACE_FULL     "ZORP_MEMTRACE_FULL"
 
 typedef struct _ZMemTraceCanary
 {
-  gint size;
-  gint neg_size;
+  gsize size;
+  gsize neg_size;
   guint32 canary[MEMTRACE_CANARY_SIZE];
 } ZMemTraceCanary;
 
 typedef struct _ZMemTraceEntry
 {
   guint32 next;
-  guint32 ptr;
-  guint32 size;
+  gpointer ptr;
+  gsize size;
   gpointer backtrace[MEMTRACE_BACKTRACE_LEN];
   gboolean deleted;
 } ZMemTraceEntry;
@@ -96,18 +101,21 @@ typedef struct _ZMemTraceHead
 #define MEMTRACE_HASH_MASK (32767 << 3)
 #define MEMTRACE_HASH_SHIFT 3
 
-/* at most this amount of blocks can be allocated at the same time 
+/**
+ * At most this amount of blocks can be allocated at the same time.
  * This preallocates MEMTRACE_HEAP_SIZE * sizeof(ZMemTraceEntry), 
- * which is 268 bytes currently. (17MB)
- */
+ * which is 268 bytes currently in i386. (17MB)
+ **/
 #define MEMTRACE_HEAP_SIZE 300000
 
 #define MEMTRACE_TEMP_HEAP_SIZE 65536
-/* If not enough 131072 */
+/**< If not enough 131072 */
+
+#define MEMTRACE_EOL ((guint32) -1)
 
 ZMemTraceHead mem_trace_hash[MEMTRACE_HASH_SIZE];
 ZMemTraceEntry mem_trace_heap[MEMTRACE_HEAP_SIZE];
-guint32 mem_trace_free_list = -1;
+guint32 mem_trace_free_list = MEMTRACE_EOL;
 guint32 mem_block_count = 0, mem_allocated_size = 0, mem_alloc_count = 0;
 gboolean mem_trace_initialized = FALSE;
 static GStaticMutex mem_trace_lock = G_STATIC_MUTEX_INIT;
@@ -122,7 +130,7 @@ gchar temp_heap[MEMTRACE_TEMP_HEAP_SIZE];
 gint temp_brk = 0;
 gint mem_trace_recurse = 0;
 
-#define TMP_ALLOCATED(ptr) (((guint32) ((char *) ptr - temp_heap)) < MEMTRACE_TEMP_HEAP_SIZE)
+#define TMP_ALLOCATED(ptr) (((unsigned int )((char *) ptr - temp_heap)) < MEMTRACE_TEMP_HEAP_SIZE)
 
 void *(*old_malloc)(size_t size);
 void (*old_free)(void *ptr);
@@ -131,6 +139,9 @@ void *(*old_calloc)(size_t nmemb, size_t size);
 
 static void z_mem_trace_printf(char *format, ...);
 
+/**
+ * Save pointers to original free, realloc, calloc and malloc.
+ **/
 static void
 z_mem_trace_init_pointers(void)
 {
@@ -144,6 +155,12 @@ z_mem_trace_init_pointers(void)
   old_malloc = dlsym(RTLD_NEXT, "malloc");
 }
 
+/**
+ * Internal function to initialize memtrace globals.
+ *
+ * Only the first call will have any effect (unless mem_trace_initialize
+ * is somehow unset again).
+ **/
 static void
 z_mem_trace_init_internal(void)
 {
@@ -151,7 +168,13 @@ z_mem_trace_init_internal(void)
   
   if (!mem_trace_initialized)
     {
+      gpointer temp_buf[10];
+
       mem_trace_initialized = TRUE;
+
+  #if defined(__AMD64__) || defined(AMD64) || defined (__x86_64__)
+      backtrace(temp_buf, 5);
+  #endif /* __amd64__ */
       
       z_mem_trace_init_pointers();
       
@@ -169,73 +192,183 @@ z_mem_trace_init_internal(void)
       
           for (i = 0; i < MEMTRACE_HEAP_SIZE; i++)
             {
-              mem_trace_heap[i].next = i+1;
+              mem_trace_heap[i].next = i + 1;
             }
-          mem_trace_heap[MEMTRACE_HEAP_SIZE - 1].next = -1;
+          mem_trace_heap[MEMTRACE_HEAP_SIZE - 1].next = MEMTRACE_EOL;
           mem_trace_free_list = 0;
       
           for (i = 0; i < MEMTRACE_HASH_SIZE; i++) 
             {
-              mem_trace_hash[i].list = -1;
-              memset(&mem_trace_hash[i].lock, 0, sizeof(GStaticMutex));
+              mem_trace_hash[i].list = MEMTRACE_EOL;
+              g_static_mutex_init(&mem_trace_hash[i].lock);
             }
         }
     }
 }
 
+/**
+ * If the ZORP_ENV_MEMTRACE environment variable says memtrace is needed,
+ * set mem_trace and set the trace file name with full path.
+ *
+ * @param[in] tracefile basename of the trace file.
+ **/
 void
 z_mem_trace_init(gchar *tracefile)
 {
-  if (getenv(ZORP_ENV_MEMTRACE))
-    mem_trace = TRUE;
+
+  z_mem_trace_init_internal();
 
   if (tracefile && mem_trace)
     {
-      g_snprintf(mem_trace_filename, sizeof(mem_trace_filename),
-                 "/var/tmp/%s", tracefile);
+      g_snprintf(mem_trace_filename, sizeof(mem_trace_filename), "/var/tmp/%s", tracefile);
+      z_mem_trace_printf("MemTrace initialized; memtrace='%s', canaries='%s', keep_deleted='%s', full_log='%s'\n",
+                YES_NO_STR(mem_trace), YES_NO_STR(mem_trace_canaries), YES_NO_STR(mem_trace_hard), YES_NO_STR(really_trace_malloc));
     }
 }
 
 static guint32 
-z_mem_trace_hash(guint32 ptr)
+z_mem_trace_hash(gpointer ptr)
 {
-  return (ptr & MEMTRACE_HASH_MASK) >> MEMTRACE_HASH_SHIFT;
+  return (guint32)((gsize)ptr & MEMTRACE_HASH_MASK) >> MEMTRACE_HASH_SHIFT;
 }
 
+#if defined(__i386__)
+
+/**
+ * Produce a backtrace from the stack into backtrace.
+ *
+ * @param[out] backtrace backtrace will be placed here.
+ *
+ * x86 version.
+ **/
 void
-z_mem_trace_bt(gpointer backtrace[])
+z_mem_trace_bt(gpointer backt[])
 {
-  /* NOTE: this is i386 specific */
   gpointer x;
-  gpointer *ebp = &x+1;
+  gpointer *ebp;
   gint i = 0;
 
-  while ((ebp > &x) && *ebp && i < MEMTRACE_BACKTRACE_LEN - 1)
+  #if defined(__i386)
+    __asm__("mov %%ebp, %0\n": "=r"(ebp));
+  #else
+  #if defined(__AMD64__) || defined(AMD64) || defined(__x86_64__)
+    __asm__("mov %%rbp, %0\n": "=r"(ebp));
+  #endif
+  #endif
+
+  while ((ebp > &x) && *ebp && *(ebp + 1) && i < MEMTRACE_BACKTRACE_LEN - 1)
     {
       gpointer value = *(ebp + 1);
       
-      backtrace[i] = value;
+      backt[i] = value;
       i++;
       ebp = *ebp;
     }
-  backtrace[i] = NULL;
+  backt[i] = NULL;
 }
 
+#else
+#if defined(__AMD64__) || defined(AMD64) || defined(__x86_64__)
+
+void
+z_mem_trace_bt(gpointer backt[])
+{
+  gpointer btrace[MEMTRACE_BACKTRACE_LEN + 1];
+  gint i = 0;
+  gint length;
+
+  length = backtrace(btrace, MEMTRACE_BACKTRACE_LEN);
+
+  while (btrace[i] && i < length && i < MEMTRACE_BACKTRACE_LEN - 1)
+    {
+      backt[i] = btrace[i];
+      i++;
+    }
+  backt[i] = NULL;
+}
+
+#else
+#if defined(__sparc__)
+
+/**
+ * Produce a backtrace from the stack into backtrace.
+ *
+ * @param[out] backt backtrace will be placed here.
+ *
+ * SPARC version.
+ **/
+void
+z_mem_trace_bt(gpointer backt[])
+{
+  gpointer x;
+  gpointer *fp;
+  gint i = 0;
+
+  __asm__("mov %%fp, %0\n": "=r"(fp));
+
+  while ((fp > &x) && *fp && i < MEMTRACE_BACKTRACE_LEN - 1)
+    {
+      gpointer value = *(fp + 16);
+      
+      backt[i] = value;
+      i++;
+      fp = *(fp + 15);
+    }
+  backt[i] = NULL;
+}
+
+#else
+
+/* Stack is not known do dummy backtrace */
+/**
+ * Produce a backtrace from the stack into backt -- except stack is unknown so this is a dummy.
+ *
+ * @param[out] backt instead of backtrace, a single NULL will be placed here.
+ *
+ * Unknown architecture version.
+ **/
+void
+z_mem_trace_bt(gpointer backt[])
+{
+  backt[0] = NULL;
+}
+#endif /* __sparc__ */
+#endif /* __x86_64__ */
+#endif /* __i386__ */
+
+/**
+ * Format backtrace into a printable string format.
+ *
+ * @param[in]  backt the backtrace
+ * @param[out] buf buffer to receive the string
+ * @param[in]  buflen size of buffer
+ *
+ * @returns a pointer to the string (same as buf)
+ **/
 static char *
-z_mem_trace_format_bt(gpointer backtrace[], gchar *buf, gint buflen)
+z_mem_trace_format_bt(gpointer backt[], gchar *buf, guint buflen)
 {
   gchar *p = buf;
   gint i, len;
+
+  p[0] = '\0';
   
-  for (i = 0; i < MEMTRACE_BACKTRACE_LEN && buflen >= 12 && backtrace[i]; i++)
+  for (i = 0; i < MEMTRACE_BACKTRACE_LEN && buflen >= (sizeof(gpointer) * 2) + 3 && backt[i]; i++)
     {
-      len = g_snprintf(buf, buflen, "%p,", backtrace[i]);
+      len = g_snprintf(buf, buflen, "%p,", backt[i]);
       buf += len;
       buflen -= len;
     }
   return p;
 }
 
+/**
+ * Write a line to the memtrace log if memtrace is enabled.
+ *
+ * @param[in] format printf-like format string
+ * 
+ * The format string and the rest of the parameters are expected to be like those of *printf.
+ **/
 static void
 z_mem_trace_printf(char *format, ...)
 {
@@ -258,6 +391,9 @@ z_mem_trace_printf(char *format, ...)
     }
 }
 
+/**
+ * Log statistics to the memtrace log.
+ **/
 void
 z_mem_trace_stats(void)
 {
@@ -279,11 +415,11 @@ z_mem_trace_dump()
         {
           ZMemTraceHead *head = &mem_trace_hash[i];
           ZMemTraceEntry *entry;
-          int cur;
+          guint32 cur;
           
           g_static_mutex_lock(&head->lock);
           cur = head->list;
-          while (cur != -1)
+          while (cur != MEMTRACE_EOL)
             {
               char backtrace_buf[MEMTRACE_BACKTRACE_BUF_LEN];
               
@@ -293,14 +429,14 @@ z_mem_trace_dump()
 
               if (mem_trace_canaries)
                 {
-                  z_mem_trace_check_canaries((gpointer)entry->ptr);
+                  z_mem_trace_check_canaries(entry->ptr);
                 }
               if (mem_trace_hard && entry->deleted)
                 {
-                  guint i;
-                  for (i = 0; i < entry->size; i++)
+                  guint j;
+                  for (j = 0; j < entry->size; j++)
                     {
-                      if (*((unsigned char *)(entry->ptr) + i) != MEMTRACE_CANARY_FILL)
+                      if (*((unsigned char *)(entry->ptr) + j) != MEMTRACE_CANARY_FILL)
                         {
                           z_mem_trace_printf("Using freed pointer; ptr=%p\n", entry->ptr);
                         }
@@ -314,11 +450,13 @@ z_mem_trace_dump()
 }
 
 /**
- * @ptr raw pointer
- * @size original size
+ * Fill areas before and after specified area with canary values.
  *
- * returns the pointer to be returned
- */
+ * @param[in] ptr raw pointer
+ * @param[in] size original size
+ *
+ * @returns the pointer to be returned
+ **/
 static gpointer
 z_mem_trace_fill_canaries(gpointer ptr, gint size)
 {
@@ -340,13 +478,12 @@ z_mem_trace_fill_canaries(gpointer ptr, gint size)
 }
 
 /**
- * @ptr user pointer
+ * Aborts the process if the canaries are touched.
  *
- * returns the pointer to be freed
+ * @param[in] ptr user pointer
  *
- * Aborts if the canaries are touched.
- * 
- */
+ * @returns the pointer to be freed
+ **/
 static gpointer
 z_mem_trace_check_canaries(gpointer ptr)
 {
@@ -360,14 +497,14 @@ z_mem_trace_check_canaries(gpointer ptr)
       
       if (p_before->size != -p_before->neg_size)
         {
-          z_mem_trace_printf("Inconsystency in canaries; ptr=%p\n", ptr);
+          z_mem_trace_printf("Inconsistency in canaries; ptr=%p\n", ptr);
           abort();
         }
       p_after = (ZMemTraceCanary *) (((gchar *) ptr) + p_before->size);
       if (p_after->size != p_before->size ||
           p_after->neg_size != p_before->neg_size)
         {
-          z_mem_trace_printf("Inconsystency in canaries; ptr=%p\n", ptr);
+          z_mem_trace_printf("Inconsistency in canaries; ptr=%p\n", ptr);
           abort();
         }
       for (i = 0; i < MEMTRACE_CANARY_SIZE; i++)
@@ -385,15 +522,16 @@ z_mem_trace_check_canaries(gpointer ptr)
 }
 
 static gboolean
-z_mem_trace_add(gpointer ptr, gint size, gpointer backtrace[])
+z_mem_trace_add(gpointer ptr, gint size, gpointer backt[])
 {
   guint32 hash, new_ndx;
   ZMemTraceEntry *new;
   ZMemTraceHead *head;
+  static time_t prev_stats = 0, now;
   
-  hash = z_mem_trace_hash((guint32) ptr);
+  hash = z_mem_trace_hash(ptr);
   g_static_mutex_lock(&mem_trace_lock);
-  if (mem_trace_free_list == -1)
+  if (mem_trace_free_list == MEMTRACE_EOL)
     {
       g_static_mutex_unlock(&mem_trace_lock);
       return FALSE;
@@ -402,8 +540,12 @@ z_mem_trace_add(gpointer ptr, gint size, gpointer backtrace[])
   mem_block_count++;
   mem_alloc_count++;
   
-  if ((mem_alloc_count % 1000) == 0)
-    z_mem_trace_stats();
+  now = time(NULL);
+  if (now != prev_stats)
+    {
+      prev_stats = now;
+      z_mem_trace_stats();
+    }
   
   mem_allocated_size += size;
   new_ndx = mem_trace_free_list;
@@ -412,9 +554,9 @@ z_mem_trace_add(gpointer ptr, gint size, gpointer backtrace[])
   
   g_static_mutex_unlock(&mem_trace_lock);
   
-  new->ptr = (guint32) ptr;
+  new->ptr = ptr;
   new->size = size;
-  memmove(new->backtrace, backtrace, sizeof(new->backtrace));
+  memmove(new->backtrace, backt, sizeof(new->backtrace));
   head = &mem_trace_hash[hash];
   
   g_static_mutex_lock(&head->lock);
@@ -427,7 +569,8 @@ z_mem_trace_add(gpointer ptr, gint size, gpointer backtrace[])
   if (really_trace_malloc)
     {
       gchar buf[1024];
-      z_mem_trace_printf("memtrace addblock; ptr='%p', size='%d', bt='%s'\n", ptr, size, z_mem_trace_format_bt(backtrace, buf, sizeof(buf)));
+      z_mem_trace_printf("memtrace addblock; ptr='%p', size='%d', bt='%s'\n",
+                         ptr, size, z_mem_trace_format_bt(backt, buf, sizeof(buf)));
     }
 
   return TRUE;
@@ -440,20 +583,20 @@ z_mem_trace_del(gpointer ptr, gpointer bt[])
   ZMemTraceHead *head;
   ZMemTraceEntry *entry;
   
-  hash = z_mem_trace_hash((guint32) ptr);
+  hash = z_mem_trace_hash(ptr);
   head = &mem_trace_hash[hash];
   
   g_static_mutex_lock(&head->lock);
   
   prev = &head->list;
   cur = head->list;
-  while (cur != -1 && mem_trace_heap[cur].ptr != (guint32) ptr)
+  while (cur != MEMTRACE_EOL && mem_trace_heap[cur].ptr != ptr)
     {
       prev = &mem_trace_heap[cur].next;
       cur = mem_trace_heap[cur].next;
     }
   
-  if (cur == -1)
+  if (cur == MEMTRACE_EOL)
     {
       g_static_mutex_unlock(&head->lock);
       
@@ -494,10 +637,10 @@ z_mem_trace_del(gpointer ptr, gpointer bt[])
 static inline guint32
 z_mem_trace_lookup_chain(gpointer ptr, ZMemTraceHead *head)
 {
-  guint32 cur = -1;
+  guint32 cur = MEMTRACE_EOL;
   
   cur = head->list;
-  while (cur != -1 && mem_trace_heap[cur].ptr != (guint32) ptr)
+  while (cur != MEMTRACE_EOL && mem_trace_heap[cur].ptr != ptr)
     {
       cur = mem_trace_heap[cur].next;
     }
@@ -513,13 +656,13 @@ z_mem_trace_getsize(gpointer ptr)
   ZMemTraceHead *head;
   
   
-  hash = z_mem_trace_hash((guint32) ptr);
+  hash = z_mem_trace_hash(ptr);
   head = &mem_trace_hash[hash];
   
   g_static_mutex_lock(&head->lock);
   cur = z_mem_trace_lookup_chain(ptr, head);
   
-  if (cur != -1)
+  if (cur != MEMTRACE_EOL)
     {
       size = mem_trace_heap[cur].size;
       g_static_mutex_unlock(&head->lock);
@@ -533,7 +676,7 @@ z_mem_trace_getsize(gpointer ptr)
 }
 
 void *
-z_malloc(size_t size, gpointer backtrace[])
+z_malloc(size_t size, gpointer backt[])
 {
   gpointer raw_ptr, user_ptr;
   gchar buf[MEMTRACE_BACKTRACE_BUF_LEN];
@@ -560,16 +703,16 @@ z_malloc(size_t size, gpointer backtrace[])
 
       if (mem_trace_hard && z_mem_trace_getsize(user_ptr) != -1)
         {
-          z_mem_trace_printf("Duplicate memory block; backtrace='%s'\n", z_mem_trace_format_bt(backtrace, buf, sizeof(buf)));
+          z_mem_trace_printf("Duplicate memory block; backtrace='%s'\n",
+                             z_mem_trace_format_bt(backtrace, buf, sizeof(buf)));
           abort();
         }
 
-      if (user_ptr && !z_mem_trace_add(user_ptr, size, backtrace))
+      if (user_ptr && !z_mem_trace_add(user_ptr, size, backt))
         {
-          gchar buf[MEMTRACE_BACKTRACE_BUF_LEN];
-      
           old_free(raw_ptr);
-          z_mem_trace_printf("Out of free memory blocks; backtrace='%s'\n", z_mem_trace_format_bt(backtrace, buf, sizeof(buf)));
+          z_mem_trace_printf("Out of free memory blocks; backtrace='%s'\n",
+                             z_mem_trace_format_bt(backt, buf, sizeof(buf)));
           z_mem_trace_stats();
           z_mem_trace_dump();
           return NULL;
@@ -581,7 +724,7 @@ z_malloc(size_t size, gpointer backtrace[])
 }
 
 void
-z_free(void *user_ptr, gpointer backtrace[])
+z_free(void *user_ptr, gpointer backt[])
 {
   gchar backtrace_buf[MEMTRACE_BACKTRACE_BUF_LEN];
   gpointer raw_ptr;
@@ -592,15 +735,16 @@ z_free(void *user_ptr, gpointer backtrace[])
   if (mem_trace)
     {
       size = z_mem_trace_getsize(user_ptr);
-      if (user_ptr && !z_mem_trace_del(user_ptr, backtrace))
+      if (user_ptr && !z_mem_trace_del(user_ptr, backt))
         {
-          z_mem_trace_printf("Trying to free a non-existing memory block; ptr=%p, backtrace='%s'\n", user_ptr, z_mem_trace_format_bt(backtrace, backtrace_buf, sizeof(backtrace_buf)));
+          z_mem_trace_printf("Trying to free a non-existing memory block; ptr=%p, backtrace='%s'\n",
+                             user_ptr, z_mem_trace_format_bt(backt, backtrace_buf, sizeof(backtrace_buf)));
           assert(0);
         }
       raw_ptr = z_mem_trace_check_canaries(user_ptr);
 
       if (size != -1)
-        memset(user_ptr, 0xcd, size);
+        memset(user_ptr, MEMTRACE_CANARY_FILL, size);
     }
   else
     {
@@ -612,10 +756,10 @@ z_free(void *user_ptr, gpointer backtrace[])
 }
 
 void  *
-z_realloc(void *user_ptr, size_t size, gpointer backtrace[])
+z_realloc(void *user_ptr, size_t size, gpointer backt[])
 {
   void *new_ptr, *raw_ptr = NULL;
-  int old_size = 0;
+  size_t old_size = 0;
   gchar buf[MEMTRACE_BACKTRACE_BUF_LEN];
   
   z_mem_trace_init_internal();
@@ -625,9 +769,10 @@ z_realloc(void *user_ptr, size_t size, gpointer backtrace[])
       if (user_ptr)
         {
           old_size = z_mem_trace_getsize(user_ptr);
-          if (old_size == -1 || !z_mem_trace_del(user_ptr, backtrace))
+          if (old_size == (size_t) -1 || !z_mem_trace_del(user_ptr, backt))
             {
-              z_mem_trace_printf("Trying to realloc a non-existing memory block; ptr=%p, size='%d', info='%s'", user_ptr, size, z_mem_trace_format_bt(backtrace, buf, sizeof(buf)));
+              z_mem_trace_printf("Trying to realloc a non-existing memory block; ptr=%p, size='%d', info='%s'",
+                                 user_ptr, size, z_mem_trace_format_bt(backt, buf, sizeof(buf)));
               assert(0);
             }
           raw_ptr = z_mem_trace_check_canaries(user_ptr);
@@ -644,7 +789,8 @@ z_realloc(void *user_ptr, size_t size, gpointer backtrace[])
         {
           /* this ptr was allocated on the temp heap, move it to real heap */
           
-          z_mem_trace_printf("reallocing space on the temp heap, moving..., ptr=%p, temp_heap=%p, diff=%d, old_size=%d\n", raw_ptr, temp_heap, (char *) raw_ptr-temp_heap, old_size);
+          z_mem_trace_printf("reallocing space on the temp heap, moving..., ptr=%p, temp_heap=%p, diff=%d, old_size=%d\n",
+                             raw_ptr, temp_heap, (char *) raw_ptr-temp_heap, old_size);
           new_ptr = old_malloc(size + mem_trace_canaries * MEMTRACE_CANARY_OVERHEAD);
           if (new_ptr)
             {
@@ -667,8 +813,8 @@ z_realloc(void *user_ptr, size_t size, gpointer backtrace[])
               /* fill_canaries doesn't touch data, only fills the canary info */
               new_ptr = z_mem_trace_fill_canaries(new_ptr, size);
               memmove(new_ptr, user_ptr, MIN(size, old_size));
-              if (old_size != -1)
-                memset(user_ptr, 0xcd, old_size);
+              if (old_size != (size_t) -1)
+                memset(user_ptr, MEMTRACE_CANARY_FILL, old_size);
             }
         }
     }
@@ -684,13 +830,13 @@ z_realloc(void *user_ptr, size_t size, gpointer backtrace[])
     }
   if (new_ptr)
     {
-      z_mem_trace_add(new_ptr, size, backtrace);
+      z_mem_trace_add(new_ptr, size, backt);
     }
   return new_ptr;
 }
 
 void *
-z_calloc(size_t nmemb, size_t size, gpointer backtrace[])
+z_calloc(size_t nmemb, size_t size, gpointer backt[])
 {
   void *user_ptr, *raw_ptr;
 
@@ -707,7 +853,7 @@ z_calloc(size_t nmemb, size_t size, gpointer backtrace[])
   if (mem_trace)
     {
       user_ptr = z_mem_trace_fill_canaries(raw_ptr, nmemb * size);
-      z_mem_trace_add(user_ptr, nmemb * size, backtrace);
+      z_mem_trace_add(user_ptr, nmemb * size, backt);
     }
   else
     {
@@ -726,61 +872,69 @@ z_calloc(size_t nmemb, size_t size, gpointer backtrace[])
 void *
 malloc(size_t size)
 {
-  gpointer backtrace[MEMTRACE_BACKTRACE_LEN];
+  gpointer backt[MEMTRACE_BACKTRACE_LEN];
 
   if (mem_trace)
-    z_mem_trace_bt(backtrace);
+    z_mem_trace_bt(backt);
   else
-    backtrace[0] = 0;
-  return z_malloc(size, backtrace);
+    backt[0] = 0;
+  return z_malloc(size, backt);
 }
 
 void
 free(void *ptr)
 {
-  gpointer backtrace[MEMTRACE_BACKTRACE_LEN];
+  gpointer backt[MEMTRACE_BACKTRACE_LEN];
   
 
   if (mem_trace)
-    z_mem_trace_bt(backtrace);
+    z_mem_trace_bt(backt);
   else
-    backtrace[0] = 0;
-  return z_free(ptr, backtrace);
+    backt[0] = 0;
+  return z_free(ptr, backt);
 }
 
 void *
 realloc(void *ptr, size_t size)
 {
-  gpointer backtrace[MEMTRACE_BACKTRACE_LEN];
+  gpointer backt[MEMTRACE_BACKTRACE_LEN];
   
 
   if (mem_trace)
-    z_mem_trace_bt(backtrace);
+    z_mem_trace_bt(backt);
   else
-    backtrace[0] = 0;
-  return z_realloc(ptr, size, backtrace);
+    backt[0] = 0;
+  return z_realloc(ptr, size, backt);
 }
 
 void *
 calloc(size_t nmemb, size_t size)
 {
-  gpointer backtrace[MEMTRACE_BACKTRACE_LEN];
+  gpointer backt[MEMTRACE_BACKTRACE_LEN];
   
 
   if (mem_trace)
-    z_mem_trace_bt(backtrace);
+    z_mem_trace_bt(backt);
   else
-    backtrace[0] = 0;
-  return z_calloc(nmemb, size, backtrace);
+    backt[0] = 0;
+  return z_calloc(nmemb, size, backt);
 }
 
 #else
 
+/**
+ * Save pointers to original free, realloc, calloc and malloc -- dummy version for when memtrace is disabled.
+ *
+ * @param memtrace_file unused
+ **/
 void 
 z_mem_trace_init(gchar *memtrace_file G_GNUC_UNUSED)
 {
 }
 
+/**
+ * Log statistics to the memtrace log -- dummy version for when memtrace is disabled.
+ **/
 void 
 z_mem_trace_stats(void)
 {
