@@ -22,12 +22,8 @@
 #include <stdlib.h>
 
 #include <ctype.h>
-/*
- * ZCharSet is a simple representation of a character set. It can be parsed
- * from a string representation where regexp like expressions can be used. 
- * Checking whether a given character is in the character set is quite fast
- * as each character is represented as a bit in a bitstring.
- */
+#include <sys/types.h>
+
 
 #define PARSE_STATE_START  0
 #define PARSE_STATE_DASH   1
@@ -35,10 +31,9 @@
 #define PARSE_STATE_ESCAPE 3
 
 /**
- * z_charset_init:
- * @self: ZCharSet instance to initialize
- *
  * Initialize a ZCharSet by setting clearing the character set bitstring.
+ *
+ * @param[out] self ZCharSet instance to initialize
  **/
 void 
 z_charset_init(ZCharSet *self)
@@ -47,11 +42,13 @@ z_charset_init(ZCharSet *self)
 }
 
 /**
- * z_charset_parse:
- * @self: ZCharSet instance, previously initialized by z_charset_init
- * @interval_str: string representation of the character set
- *
  * This function parses an character set from its string representation.
+ *
+ * @param[in] self ZCharSet instance, previously initialized by z_charset_init
+ * @param[in] interval_str string representation of the character set
+ *
+ * @returns true if interval_str parsed correctly (the parser didn't get into a
+ *      nonexistent state and the parser returned to its starting state at the end)
  **/
 gboolean 
 z_charset_parse(ZCharSet *self, gchar *interval_str)
@@ -84,6 +81,7 @@ z_charset_parse(ZCharSet *self, gchar *interval_str)
               i++;
             }
           break;
+          
         case PARSE_STATE_DASH:
           if (interval_str[i] == '\\' && old_state != PARSE_STATE_ESCAPE)
             {
@@ -102,6 +100,7 @@ z_charset_parse(ZCharSet *self, gchar *interval_str)
                 i--;
             }
           break;
+          
         case PARSE_STATE_END:
           if (interval_str[i] == '\\' && old_state != PARSE_STATE_ESCAPE)
             {
@@ -121,15 +120,16 @@ z_charset_parse(ZCharSet *self, gchar *interval_str)
               old_state = PARSE_STATE_END;
             }
           break;
+          
         case PARSE_STATE_ESCAPE:
           z_cp();
           i++;
           state = old_state;
           old_state = PARSE_STATE_ESCAPE;
           break;
+
         default:
-          z_leave();
-          return FALSE;
+          z_return(FALSE);
         }
     }
 
@@ -140,25 +140,17 @@ z_charset_parse(ZCharSet *self, gchar *interval_str)
       state = PARSE_STATE_START;
     }
   
-  if (state == PARSE_STATE_START)
-    {
-      z_leave();
-      return TRUE;
-    }
-  
-  z_leave();
-  return FALSE;
+  z_return(state == PARSE_STATE_START);
 }
 
 /**
- * z_charset_is_string_valid:
- * @self: ZCharSet instance
- * @str: string to check
- * @len: string length
- *
  * This function checks whether the given string contains valid characters only.
  *
- * Returns: when the string contained valid characters only
+ * @param[in] self ZCharSet instance
+ * @param[in] str string to check
+ * @param[in] len string length
+ *
+ * @returns if the string contains valid characters only
  **/
 gboolean 
 z_charset_is_string_valid(ZCharSet *self, gchar *str, gint len)
@@ -177,15 +169,15 @@ z_charset_is_string_valid(ZCharSet *self, gchar *str, gint len)
 }
 
 /**
- * g_string_assign_len:
- * @s: GString instance
- * @val: string pointer
- * @len: length of string
- *
  * This function assigns the given string/length value to the specified GString.
+ *
+ * @param[in] s GString instance
+ * @param[in] val string pointer
+ * @param[in] len length of string
+ *
  * This function should be defined in GLib, however no such function exists.
  *
- * Returns: the GString instance 
+ * @returns the GString instance 
  **/
 GString *
 g_string_assign_len(GString *s, gchar *val, gint len)
@@ -197,87 +189,201 @@ g_string_assign_len(GString *s, gchar *val, gint len)
 }
 
 /**
- * z_data_dump:
- * @session_id: session id to be used for the log messages
- * @buf: buffer
- * @len: buffer length
+ * Compares t1 and t2
  *
+ * @param[in] t1 time value t1
+ * @param[in] t2 time value t2
+ *
+ * @returns -1 if t1 is earlier than t2
+ * @returns 0 if t1 equals t2
+ * @returns 1 if t1 is later than t2
+ **/
+gint
+g_time_val_compare(const GTimeVal *t1, const GTimeVal *t2)
+{
+  g_assert(t1);
+  g_assert(t2);
+  if (t1->tv_sec < t2->tv_sec)
+    return -1;
+  else if (t1->tv_sec > t2->tv_sec)
+    return 1;
+  else if (t1->tv_usec < t2->tv_usec)
+    return -1;
+  else if (t1->tv_usec > t2->tv_usec)
+    return 1;
+  return 0;
+}
+
+/**
+ * Calculates the time difference between t1 and t2 in microseconds.
+ *
+ * @param[in] t1 time value t1
+ * @param[in] t2 time value t2
+ *
+ * The result is positive if t1 is later than t2.
+ *
+ * @returns Time difference in microseconds
+ **/
+glong
+g_time_val_diff(const GTimeVal *t1, const GTimeVal *t2)
+{
+  g_assert(t1);
+  g_assert(t2);
+  return (t1->tv_sec - t2->tv_sec) * G_USEC_PER_SEC + (t1->tv_usec - t2->tv_usec);
+}
+
+/**
+ * g_timeval_subtract:
+ * @result: the value of x - y
+ * @x:
+ * @y:
+ *
+ * Subtract y from x. The value of x is always greater or equals to y.
+ * The result is non-negative.
+ *
+ * It cannot be used to calculate a negative result.
+ *
+ */
+void
+g_time_val_subtract(GTimeVal *result, const GTimeVal *x, const GTimeVal *y)
+{
+  result->tv_usec = x->tv_usec;
+  result->tv_sec = x->tv_sec;
+
+  if (x->tv_usec < y->tv_usec)
+    {
+      result->tv_usec += 1000000;
+      result->tv_sec -= 1;
+    }
+  result->tv_usec -= y->tv_usec;
+  result->tv_sec -= y->tv_sec;
+}
+
+/**
+ * Produces one line of hex dump of the specified part buf in line.
+ *
+ * @param[out] line where the result will appear
+ * @param[in]  linelen allocated length of line
+ * @param[in]  i index in buf to dump from
+ * @param[in]  buf raw binary data to dump
+ * @param[in]  len length of buf
+ *
+ * This function will hexdump up to 16 bytes of buf on one line.
+ * This will be followed with a dump of the characters where
+ * unprintable characters will be replaced with '.'
+ *
+ * @returns The number of characters that were actually dumped on this line
+ **/
+static guint
+z_hexdump(gchar *line, guint linelen, guint i, const char *buf, guint len)
+{
+  guint j;
+  char *end = line;
+
+  for (j = 0; j < 16 && (i + j < len); j++)
+    {
+      g_snprintf(end, linelen - (end - line), "%02X ", (unsigned char) buf[i+j]);
+      end += 3;
+    }
+  for (; j < 16; j++)
+    {
+      g_snprintf(end, linelen - (end - line), "   ");
+      end += 3;
+    }
+  g_snprintf(end, linelen - (end - line), " ");
+  end++;
+  for (j = 0; j < 16 && (i + j < len) && linelen > (guint)(end - line); j++)
+    {
+      *end = isprint(buf[i + j]) ? buf[i + j] : '.';
+      end++;
+    }
+  *end='\0';
+
+  return j;
+}
+
+
+/**
  * This function generates hexdumps of the specified buffer to the system
  * log using z_log().
+ *
+ * @param[in] session_id session id to be used for the log messages
+ * @param[in] class log message class
+ * @param[in] level log message verbosity level
+ * @param[in] buf buffer
+ * @param[in] len buffer length
+ *
+ * Used internally by z_log_data_dump().
  **/
 void 
-z_data_dump(char *session_id, const char *buf, guint len)
+z_format_data_dump(const gchar *session_id, const char *class, gint level, const void *buf, guint len)
 {
-  guint i, j;
+  guint i, offs;
   gchar line[1024];
   
   i = 0;
   while (i < len)
     {
-      char *end = line;
-      
-      for (j = 0; j < 16 && (i + j < len); j++)
-        {
-          g_snprintf(end, sizeof(line) - (end - line), "%02X ", (unsigned char) buf[i+j]);
-          end += 3;
-        }
-      g_snprintf(end, sizeof(line) - (end - line), " ");
-      end++;
-      for (j = 0; j < 16 && (i + j < len) && sizeof(line) > (guint)(end - line); j++)
-        {
-          *end = buf[i + j] > 31 ? buf[i + j] : '.';
-          end++;
-        }
-      *end='\0';
-          
-      i += j;
-      
+      offs = i;
+      i += z_hexdump(line, sizeof(line), i, buf, len);
       /*NOLOG*/
-      z_log(session_id, CORE_DUMP, 9, "data line: %s", line);
+      z_log((gchar *)session_id, class, level, "data line 0x%04x: %s", offs, line);
     }
 }
 
 /**
- * z_format_data_dump:
- * @session_id: session id to be used for the log messages
- * @buf: buffer
- * @len: buffer length
- *
- * This function generates hexdumps of the specified buffer to the system
+ * This function generates textual dumps of the specified buffer to the system
  * log using z_log().
+ *
+ * @param[in] session_id session id to be used for the log messages
+ * @param[in] class log message class
+ * @param[in] level log message verbosity level
+ * @param[in] buf buffer
+ * @param[in] len buffer length
+ *
+ * Used internally by z_log_text_dump().
  **/
-void
-z_format_data_dump(gchar *session_id, char *class, gint level, const char *buf, guint len)
+void 
+z_format_text_dump(const gchar *session_id, const char *class, gint level, const void *buf, guint len)
 {
-  guint i, j;
+  guint i, nl;
   gchar line[1024];
-
-  i = 0;
-  while (i < len)
+  const gchar *bufc = buf;
+ 
+  while (len > 0)
     {
-      char *end = line;
-
-      for (j = 0; j < 16 && (i + j < len); j++)
-        {
-          g_snprintf(end, sizeof(line) - (end - line), "%02X ", (unsigned char) buf[i+j]);
-          end += 3;
-        }
-      g_snprintf(end, sizeof(line) - (end - line), " ");
-      end++;
-      for (j = 0; j < 16 && (i + j < len) && sizeof(line) > (guint)(end - line); j++)
-        {
-          *end = buf[i + j] > 31 ? buf[i + j] : '.';
-          end++;
-        }
-      *end='\0';
-
-      i += j;
-
+      for (nl = 0; (nl < len) && bufc[nl] && (bufc[nl] != '\r') && (bufc[nl] != '\n'); nl++)
+        ;
+      i = (nl < len) ? nl : nl - 1; /* # of chars to dump */
+      if (i >= sizeof(line))
+        i = sizeof(line) - 1;
+      memcpy(line, bufc, i);
+      line[i] = '\0';
       /*NOLOG*/
-      z_log(session_id, class, level, "data line: %s", line);
+      z_log((gchar *)session_id, class, level, "text line: %s", line);
+      bufc += nl;
+      len -= nl;
+      if ((len > 0) && (bufc[0] == '\r'))
+        {
+          bufc++;
+          len--;
+        }
+      if ((len > 0) && (bufc[0] == '\n'))
+        {
+          bufc++;
+          len--;
+        }
     }
 }
 
+/**
+ * Escapes spaces to %_ and %-s to %% in s and returns the result.
+ *
+ * @param[in] s the string to escape characters in
+ * @param[in] len length of s
+ *
+ * @returns the escaped string
+ **/
 gchar *
 z_str_escape(const gchar *s, gint len)
 {
@@ -285,12 +391,9 @@ z_str_escape(const gchar *s, gint len)
   gint i = 0, j = 0;;
   
   z_enter();
-  
   if (len < 0)
     len = strlen(s) + 1;
-  
   res = g_new0(gchar, len * 2);
-  
   while (i < len && s[i] != '\0')
     {
       switch (s[i])
@@ -310,11 +413,17 @@ z_str_escape(const gchar *s, gint len)
 	}
       i++;
     }
-  
-  z_leave();
-  return res;
+  z_return(res);
 }
 
+/**
+ * Undoes the escaping done by z_str_escape() on s and returns the result.
+ *
+ * @param[in] s the string to unescape characters in
+ * @param[in] len length of s
+ *
+ * @returns the compressed string
+ **/
 gchar *
 z_str_compress(const gchar *s, gint len)
 {
@@ -322,12 +431,9 @@ z_str_compress(const gchar *s, gint len)
   gint i = 0, j = 0;;
   
   z_enter();
-  
   if (len < 0)
     len = strlen(s) + 1;
-  
   res = g_new0(gchar, len);
-  
   while (i < len && s[i] != '\0')
     {
       if (s[i] == '%' && s[i+1] == '%')
@@ -346,21 +452,20 @@ z_str_compress(const gchar *s, gint len)
 	}
       i++;
     }
-  
-  z_leave();
-  return res;
+  z_return(res);
 }
 
 /**
- * z_port_enabled:
- * @port_range: port range specification
- * @port: port number to check
+ * Parse port range and check if the port number is valid.
+ *
+ * @param[in] port_range port range specification
+ * @param[in] port port number to check
  *
  * This function parses the given port_range and returns TRUE to indicate
  * whether the specified port number is valid. The port specification is in
  * the format: port1[-port2]([,port1[-port2]])*
  *
- * Returns: TRUE when the port number is valid
+ * @returns TRUE when the port number is valid
  **/
 gboolean
 z_port_enabled(gchar *port_range, guint port)
@@ -405,11 +510,10 @@ z_port_enabled(gchar *port_range, guint port)
 }
 
 /**
- * z_casestr_equal:
- * @k1: key1 to compare
- * @k2: key2 to compare
- * 
  * GLib hashtable equal function for case insensitive hashtables.
+ *
+ * @param[in] k1 key1 to compare
+ * @param[in] k2 key2 to compare
  **/
 gboolean
 z_casestr_equal(gconstpointer k1, gconstpointer k2)
@@ -418,10 +522,9 @@ z_casestr_equal(gconstpointer k1, gconstpointer k2)
 }
 
 /**
- * z_casestr_hash:
- * @key:
- *
  * GLib hashtable hash function for case insensitive hashtables.
+ *
+ * @param[in] key
  **/
 guint
 z_casestr_hash(gconstpointer key)
@@ -436,25 +539,20 @@ z_casestr_hash(gconstpointer key)
   return h;
 }
 
-
-#define ON_OFF_STR(x) (x ? "on" : "off")
-
 /**
- * z_zorplib_version_info:
- * 
  * This function returns a static character string which includes version
  * information and compilation settings. This function is _NOT_ reentrant.
  *
- * Returns: a static string of the version information to be displayed to
+ * @returns a static string of the version information to be displayed to
  * the user
  **/
-gchar *
-z_zorplib_version_info(void)
+const gchar *
+z_libzorpll_version_info(void)
 {
-  static gchar buf[128];
+  static gchar buf[512];
   
   g_snprintf(buf, sizeof(buf),
-             "Zorplib %s\n"
+             "libzorpll %s\n"
              "Revision: %s\n"
              "Compile-Date: %s %s\n"
              "Trace: %s\n"
@@ -472,3 +570,54 @@ z_zorplib_version_info(void)
   return buf;
 }
 
+/* NOTE: we deliberately do not define this function in a header file as we don't want these public */
+void z_thread_add_option_group(GOptionContext *ctx);
+void z_process_add_option_group(GOptionContext *ctx);
+void z_log_add_option_group(GOptionContext *ctx);
+
+/**
+ * This function put the library specific options to the program
+ * option context.
+ *
+ * @param[in] ctx A #GoptionContext container for the options
+ * @param[in] disable_groups Exclude these groups.
+ **/
+void
+z_libzorpll_add_option_groups(GOptionContext *ctx, guint disable_groups)
+{
+#ifndef G_OS_WIN32  
+  /* we don't have process API on Windows */
+  if ((disable_groups & Z_OG_PROCESS) == 0)
+    z_process_add_option_group(ctx);
+#endif
+
+  if ((disable_groups & Z_OG_THREAD) == 0)
+    z_thread_add_option_group(ctx);
+
+  if ((disable_groups & Z_OG_LOG) == 0)
+    z_log_add_option_group(ctx);
+}
+
+#ifndef HAVE_LOCALTIME_R
+/**
+ * localtime_r replacement in case we don't have one.
+ *
+ * @param[in]  timep time in seconds since epoch form to convert
+ * @param[out] result tm struct to convert date/timestamp to
+ *
+ * localtime_r is implemented by locking and calling localtime.
+ *
+ * @returns a pointer to result
+ **/
+struct tm *
+localtime_r(const time_t *timep, struct tm *result)
+{
+  static GStaticMutex mutex = G_STATIC_MUTEX_INIT;
+
+  g_static_mutex_lock (&mutex);
+  memcpy(result, localtime(timep), sizeof(struct tm));
+  g_static_mutex_unlock (&mutex);
+  return result;
+}
+#endif
+
