@@ -29,27 +29,24 @@
 #include <stdlib.h>
 #include <string.h>
 
-#ifdef G_OS_WIN32
-#  include <winsock2.h>
-#else
-#  include <sys/socket.h>
-#  include <netinet/in.h>
-#  include <sys/un.h>
-#  include <arpa/inet.h>
+#ifndef G_OS_WIN32
+# include <netdb.h>
 #endif
 
 static ZSockAddrFuncs inet_range_sockaddr_funcs;
 
+#ifdef G_OS_WIN32
+#endif
+
 /**
- * z_inet_ntoa:
- * @buf: store result in this buffer
- * @bufsize: the available space in buf
- * @a: address to convert.
- *
  * Thread friendly version of inet_ntoa(), converts an IP address to
  * human readable form.
  *
- * Returns: the address of buf
+ * @param[out] buf store result in this buffer
+ * @param[in]  bufsize the available space in buf
+ * @param[in]  a address to convert.
+ *
+ * @returns the address of buf
  **/
 gchar *
 z_inet_ntoa(gchar *buf, size_t bufsize, struct in_addr a)
@@ -65,14 +62,13 @@ z_inet_ntoa(gchar *buf, size_t bufsize, struct in_addr a)
 }
 
 /**
- * z_inet_aton:
- * @buf: buffer to parse
- * @a: the parsed address
- *
  * This function takes an IP address in text representation, parses it and 
- * returns it in @a.
+ * returns it in a.
+ *
+ * @param[in]  buf buffer to parse
+ * @param[out] a the parsed address
  * 
- * Returns: TRUE to indicate that parsing was succesful.
+ * @returns TRUE to indicate that parsing was successful.
  **/
 gboolean
 z_inet_aton(const gchar *buf, struct in_addr *a)
@@ -90,33 +86,36 @@ z_inet_aton(const gchar *buf, struct in_addr *a)
 
 
 /**
- * z_sockaddr_new:
- * @sa: libc sockaddr * pointer to convert
- * @salen: size of sa
- * 
  * General function to allocate and initialize a ZSockAddr structure,
  * and convert a libc style sockaddr * pointer to our representation.
  *
- * Returns: new ZSockAddr instance
+ * @param[in] sa libc sockaddr * pointer to convert
+ * @param[in] salen size of sa
+ *
+ * @returns new ZSockAddr instance
  **/
 ZSockAddr *
-z_sockaddr_new(struct sockaddr *sa, int salen)
+z_sockaddr_new(struct sockaddr *sa, gsize salen)
 {
   z_enter();
   switch (sa->sa_family)
     {
+#ifndef G_OS_WIN32
+    case AF_INET6:
+      if (salen >= sizeof(struct sockaddr_in6))
+        z_return(z_sockaddr_inet6_new2((struct sockaddr_in6 *) sa));
+      break;
+      
+#endif
     case AF_INET:
-      if (salen != sizeof(struct sockaddr_in))
-        {
-          z_leave();
-          return NULL;
-        }
-      z_leave();
-      return z_sockaddr_inet_new2((struct sockaddr_in *) sa);
+      if (salen == sizeof(struct sockaddr_in))
+        z_return(z_sockaddr_inet_new2((struct sockaddr_in *) sa));
+      break;
+      
 #ifndef G_OS_WIN32
     case AF_UNIX:
-      z_leave();
-      return z_sockaddr_unix_new2((struct sockaddr_un *) sa, salen);
+      /* NOTE: the sockaddr_un structure might be less than struct sockaddr_un */
+      z_return(z_sockaddr_unix_new2((struct sockaddr_un *) sa, salen));
 #endif
     default:
       /*LOG
@@ -124,22 +123,20 @@ z_sockaddr_new(struct sockaddr *sa, int salen)
         unsupported address family. Please report this error to the Zorp QA team.
        */
       z_log(NULL, CORE_ERROR, 3, "Unsupported socket family in z_sockaddr_new(); family='%d'", sa->sa_family);
-      z_leave();
-      return NULL;
+      z_return(NULL);
     }
-  z_leave();
+  z_return(NULL);
 }
 
 /**
- * z_sockaddr_format:
- * @a: instance pointer of a ZSockAddr
- * @text: destination buffer
- * @n: the size of text
- *
  * Format a ZSockAddr into human readable form, calls the format
  * virtual method of ZSockAddr.
  *
- * Returns: text is filled with a human readable representation of a, and a
+ * @param[in]  a instance pointer of a ZSockAddr
+ * @param[out] text destination buffer
+ * @param[in]  n the size of text
+ *
+ * @returns text is filled with a human readable representation of a, and a
  * pointer to text is returned.
  **/
 gchar *
@@ -149,50 +146,59 @@ z_sockaddr_format(ZSockAddr *a, gchar *text, gulong n)
 }
 
 /**
- * z_sockaddr_ref:
- * a: pointer to ZSockAddr instance
- * 
+ * This function compares two ZSockAddr structures whether their _value_ is
+ * equal, e.g.\ same IP/port.
+ *
+ * @param[in] a first sockaddr to compare
+ * @param[in] b second sockaddr to compare
+ *
+ * @returns TRUE if they're equal
+ **/
+gboolean
+z_sockaddr_equal(ZSockAddr *a, ZSockAddr *b)
+{
+  return a->sa.sa_family == b->sa.sa_family && a->sa_funcs->sa_equal(a, b);
+}
+
+/**
  * Increment the reference count of a ZSockAddr instance.
  *
- * Returns: the same instance
+ * @param[in] a pointer to ZSockAddr instance
+ *
+ * @returns the same instance
  **/
 ZSockAddr *
 z_sockaddr_ref(ZSockAddr *a)
 {
   if (a)
-    a->refcnt++;
+    z_refcount_inc(&a->refcnt);
   return a;
 }
 
 /**
- * z_sockaddr_unref:
- * @a: ZSockAddr instance
- *
  * Decrement the reference count of a ZSockAddr instance, and free if
  * the refcnt reaches 0.
+ *
+ * @param[in] a ZSockAddr instance
  **/
 void 
 z_sockaddr_unref(ZSockAddr *a)
 {
-  if (a) 
+  if (a && z_refcount_dec(&a->refcnt)) 
     {
-      g_assert(a->refcnt > 0);
-      if (--a->refcnt == 0)
-        {
-          if (!a->sa_funcs->freefn)
-            g_free(a);
-          else
-            a->sa_funcs->freefn(a);
-        }
+      if (!a->sa_funcs->freefn)
+        g_free(a);
+      else
+        a->sa_funcs->freefn(a);
     }
 }
 
 /* AF_INET socket address */
 
-/*
+/**
  * ZSockAddrInet is an implementation of the ZSockAddr interface,
  * encapsulating an IPv4 host address and port number.
- */
+ **/
 typedef struct _ZSockAddrInet
 {
   gint refcnt;
@@ -200,21 +206,20 @@ typedef struct _ZSockAddrInet
   ZSockAddrFuncs *sa_funcs;
   int salen;
   struct sockaddr_in sin;
-  gpointer options;
-  guint options_length;
 } ZSockAddrInet;
 
 /**
- * z_sockaddr_inet_bind_prepare:
- * @sock: fd of the socket to prepare for bind
- * @addr: address where we are binding to
- * @sock_flags: socket flags (ZSF_*)
- *
  * This function is used as a bind_prepare callback for ZSockAddrInet
- * addresses. It currently enables SO_REUSEADDR to permit concurrent
+ * addresses.
+ *
+ * @param[in] sock fd of the socket to prepare for bind
+ * @param     addr address where we are binding to (unused)
+ * @param[in] sock_flags socket flags (ZSF_*)
+ *
+ * It currently enables SO_REUSEADDR to permit concurrent
  * access to the same socket.
  *
- * Returns: GIOStatus to indicate success/failure
+ * @returns GIOStatus to indicate success/failure
  **/
 static GIOStatus
 z_sockaddr_inet_bind_prepare(int sock, ZSockAddr *addr G_GNUC_UNUSED, guint32 sock_flags)
@@ -232,14 +237,14 @@ z_sockaddr_inet_bind_prepare(int sock, ZSockAddr *addr G_GNUC_UNUSED, guint32 so
 }
 
 /**
- * z_sockaddr_inet_clone:
- * @addr: address to clone
- * @wildcard: whether the port information is important
+ * This function is the clone callback for ZSockAddrInet sockets.
  *
- * This function is the clonse callback for ZSockAddrInet sockets. It copies
- * the sockaddr contents and nulls port if @wildcard is TRUE.
+ * @param[in] addr address to clone
+ * @param[in] wildcard whether the port information is important
  *
- * Returns: the cloned address
+ * It copies the sockaddr contents and nulls port if wildcard is TRUE.
+ *
+ * @returns the cloned address
  **/
 static ZSockAddr *
 z_sockaddr_inet_clone(ZSockAddr *addr, gboolean wildcard)
@@ -255,15 +260,14 @@ z_sockaddr_inet_clone(ZSockAddr *addr, gboolean wildcard)
 }
 
 /**
- * z_sockaddr_inet_format:
- * @addr: ZSockAddrInet instance
- * @text: buffer to format this address into
- * @n: size of text
- *
  * This function is the format callback for ZSockAddrInet socket addresses
  * and returns the human readable representation of the IPv4 address.
  *
- * Returns: the start of the buffer
+ * @param[in]  addr ZSockAddrInet instance
+ * @param[out] text buffer to format this address into
+ * @param[in]  n size of text
+ *
+ * @returns the start of the buffer
  **/
 static gchar *
 z_sockaddr_inet_format(ZSockAddr *addr, gchar *text, gulong n)
@@ -278,39 +282,57 @@ z_sockaddr_inet_format(ZSockAddr *addr, gchar *text, gulong n)
 }
 
 /**
- * z_sockaddr_inet_free:
- * @addr: ZSockAddrInet instance
+ * This function is the free callback fro ZSockAddrInet sockets.
  *
- * This function is the free callback fro ZSockAddrInet sockets. It frees 
- * ZSockAddrInet specific information and the address structure itself.
+ * @param[in] addr ZSockAddrInet instance
+ *
+ * It frees ZSockAddrInet specific information and the address structure itself.
  **/
 void
 z_sockaddr_inet_free(ZSockAddr *addr)
 {
-  ZSockAddrInet *self = (ZSockAddrInet *) addr;
-
-  if (self->options)
-    g_free(self->options);
   g_free(addr);
 }
 
+/**
+ * Checks whether two addresses of the AF_INET family are equal.
+ *
+ * @param[in] addr the first address
+ * @param[in] o the second address
+ *
+ * @returns TRUE if both the IP address and port are equal.
+ **/
+static gboolean
+z_sockaddr_inet_equal(ZSockAddr *addr, ZSockAddr *o)
+{
+  ZSockAddrInet *self = (ZSockAddrInet *) addr;
+  ZSockAddrInet *other = (ZSockAddrInet *) o;
+  
+  g_assert(self->sin.sin_family == AF_INET);
+  g_assert(other->sin.sin_family == AF_INET);
+  return self->sin.sin_addr.s_addr == other->sin.sin_addr.s_addr && self->sin.sin_port == other->sin.sin_port;
+}
+
+/**
+ * ZSockAddrInet virtual methods.
+ **/
 static ZSockAddrFuncs inet_sockaddr_funcs = 
 {
   z_sockaddr_inet_bind_prepare,
   NULL,
   z_sockaddr_inet_clone,
   z_sockaddr_inet_format,
-  z_sockaddr_inet_free
+  z_sockaddr_inet_free,
+  z_sockaddr_inet_equal
 };
 
 /**
- * z_sockaddr_inet_new:
- * @ip: text representation of an IP address
- * @port: port number in host byte order
- *
  * This constructor allocates and initializes an IPv4 socket address.
  *
- * Returns: the new instance
+ * @param[in] ip text representation of an IP address
+ * @param[in] port port number in host byte order
+ *
+ * @returns the new instance
  **/
 ZSockAddr *
 z_sockaddr_inet_new(const gchar *ip, guint16 port)
@@ -337,13 +359,12 @@ z_sockaddr_inet_new(const gchar *ip, guint16 port)
 
 
 /**
- * z_sockaddr_inet_new2:
- * @sinaddr: the sockaddr_in structure to convert
- *
  * Alternate constructor for ZSockAddrInet which takes a (struct
  * sockaddr_in) structure instead of an ip+port.
  *
- * Returns: the allocated instance.
+ * @param[in] sinaddr the sockaddr_in structure to convert
+ *
+ * @returns the allocated instance.
  **/
 ZSockAddr *
 z_sockaddr_inet_new2(struct sockaddr_in *sinaddr)
@@ -359,50 +380,45 @@ z_sockaddr_inet_new2(struct sockaddr_in *sinaddr)
   return (ZSockAddr *) self;
 }
 
+#ifndef  G_OS_WIN32
 /**
- * z_sockaddr_inet_get_address:
- * @s: ZSockAddrInet instance
+ * Alternate constructor for ZSockAddrInet which takes a hostname+port
+ * instead of a struct sockaddr_in.
  *
- * This ZSockAddrInet specific function returns the address part of the
- * address.
+ * @param[in] hostname name of the host to resolve
+ * @param[in] port port number in host byte order
+ *
+ * @returns the allocated instance.
  **/
-guint 
-z_sockaddr_inet_get_address(ZSockAddr *s)
+ZSockAddr *
+z_sockaddr_inet_new_hostname(const gchar *hostname, guint16 port)
 {
-  ZSockAddrInet *self = (ZSockAddrInet *) s;
+  struct hostent hes, *he;
+  char hostbuf[1024];
+  char buf[32];
+  int err = 0;
+  int rc;
+  ZSockAddr *saddr = NULL;
 
-  g_assert(z_sockaddr_inet_check(s));
+  rc = gethostbyname_r(hostname, &hes, hostbuf, sizeof(hostbuf), &he, &err);
 
-  return self->sin.sin_addr.s_addr;
+  if (rc == 0 && he && he->h_addr_list[0])
+    {
+      z_inet_ntoa(buf, sizeof(buf), *((struct in_addr *) he->h_addr_list[0]));
+      saddr = z_sockaddr_inet_new(buf, port);
+    }
+
+  return saddr;
 }
+#endif
 
 /**
- * z_sockaddr_inet_get_port:
- * @s: ZSockAddrInet instance
- *
- * This ZSockAddrInet specific function returns the port part of the
- * address.
- * 
- * Returns: the port in host byte order
- * 
- **/
-guint 
-z_sockaddr_inet_get_port(ZSockAddr *s)
-{
-  ZSockAddrInet *self = (ZSockAddrInet *) s;
-
-  g_assert(z_sockaddr_inet_check(s));
-  return ntohs(self->sin.sin_port);
-}
-
-/**
- * z_sockaddr_inet_check:
- * @s: ZSockAddr instance
- *
  * This function checks whether the given ZSockAddr instance is in fact a
  * ZSockAddrInet instance.
  *
- * Returns: TRUE if @s is a ZSockAddrInet
+ * @param[in] s ZSockAddr instance
+ *
+ * @returns TRUE if s is a ZSockAddrInet
  **/
 gboolean 
 z_sockaddr_inet_check(ZSockAddr *s)
@@ -410,37 +426,14 @@ z_sockaddr_inet_check(ZSockAddr *s)
   return (s->sa_funcs == &inet_sockaddr_funcs) || (s->sa_funcs == &inet_range_sockaddr_funcs);
 }
 
-#if ZORPLIB_ENABLE_IPOPTIONS
-void
-z_sockaddr_inet_get_options(ZSockAddr *s, gpointer *options, guint *options_length)
-{
-  ZSockAddrInet *self = (ZSockAddrInet *) s;
-  
-  *options = self->options;
-  *options_length = self->options_length;
-}
-
-void
-z_sockaddr_inet_set_options(ZSockAddr *s, gpointer options, guint options_length)
-{
-  ZSockAddrInet *self = (ZSockAddrInet *) s;
-  
-  if (self->options)
-    g_free(self->options);
-  
-  self->options = g_malloc0(options_length);
-  memcpy(self->options, options, options_length);
-  self->options_length = options_length;
-}
-#endif
-
-
 /*+ Similar to ZSockAddrInet, but binds to a port in the given range +*/
 
-/* 
- * NOTE: it is assumed that it is safe to cast from ZSockAddrInetRange to
+/**
+ * Class to store an IP address with a port range.
+ *
+ * @note it is assumed that it is safe to cast from ZSockAddrInetRange to
  * ZSockAddrInet
- */
+ **/
 typedef struct _ZSockAddrInetRange
 {
   gint refcnt;
@@ -448,19 +441,17 @@ typedef struct _ZSockAddrInetRange
   ZSockAddrFuncs *sa_funcs;
   int salen;
   struct sockaddr_in sin;
-  gpointer options;
-  guint options_length;
   guint16 min_port, max_port, last_port;
 } ZSockAddrInetRange;
 
 /**
- * z_sockaddr_inet_range_bind:
- * @sock: socket to bind
- * @a: address to bind to
- * @sock_flags: socket flags (ZSF_*)
+ * This function is the bind callback of ZSockAddrInetRange.
  *
- * This function is the bind callback of ZSockAddrInetRange. It tries to
- * allocate a port in the specified range.
+ * @param[in] sock socket to bind
+ * @param[in] a address to bind to
+ * @param[in] sock_flags socket flags (ZSF_*)
+ *
+ * It tries to allocate a port in the specified range.
  *
  * Returns: a GIOStatus value to indicate failure/success
  **/
@@ -506,19 +497,27 @@ z_sockaddr_inet_range_bind(int sock, ZSockAddr *a, guint32 sock_flags)
           return G_IO_STATUS_NORMAL;
         }
     }
+
+  /*LOG
+    This message reports that the SockAddRinetRange could not find any
+    free port in the given range.
+   */
+  z_log(NULL, CORE_ERROR, 3,
+        "SockAddrInetRange, could not find free port to bind; min_port='%d', max_port='%d'",
+        self->min_port, self->max_port);
   self->last_port = self->min_port;
   return G_IO_STATUS_ERROR;
 }
 
 /**
- * z_sockaddr_inet_range_clone:
- * @addr: address to clone
- * @wildcard: whether the port information is important
+ * This function is the clone callback for ZSockAddrInetRange sockets.
  *
- * This function is the clone callback for ZSockAddrInetRange sockets. It
- * copies the sockaddr contents.
+ * @param[in] addr address to clone
+ * @param     wildcard_clone whether the port information is important (unused)
  *
- * Returns: the cloned address
+ * It copies the sockaddr contents.
+ *
+ * @returns the cloned address
  **/
 static ZSockAddr *
 z_sockaddr_inet_range_clone(ZSockAddr *addr, gboolean wildcard_clone G_GNUC_UNUSED)
@@ -539,6 +538,9 @@ z_sockaddr_inet_range_clone(ZSockAddr *addr, gboolean wildcard_clone G_GNUC_UNUS
   return (ZSockAddr *) self;
 }
 
+/**
+ * ZSockAddrInetRange virtual methods.
+ **/
 static ZSockAddrFuncs inet_range_sockaddr_funcs = 
 {
   NULL,
@@ -546,37 +548,56 @@ static ZSockAddrFuncs inet_range_sockaddr_funcs =
   z_sockaddr_inet_range_clone,
   z_sockaddr_inet_format,
   z_sockaddr_inet_free,
+  z_sockaddr_inet_equal
 };
 
 /**
- * z_sockaddr_inet_range_new:
- * @ip: ip address in text form
- * @min_port: minimal port to bind to
- * @max_port: maximum port to bind to
- *
- * This constructure creates a new ZSockAddrInetRange instance with the
+ * This constructor creates a new ZSockAddrInetRange instance with the
  * specified arguments.
+ *
+ * @param[in] ip ip address in text form
+ * @param[in] min_port minimal port to bind to
+ * @param[in] max_port maximum port to bind to
+ *
+ * @returns the new ZSockAddrInetRange instance
  **/
 ZSockAddr *
 z_sockaddr_inet_range_new(const gchar *ip, guint16 min_port, guint16 max_port)
 {
-  ZSockAddrInetRange *self;
   struct in_addr netaddr;
 
   if (!z_inet_aton(ip, &netaddr))
     return NULL;
   
+  return z_sockaddr_inet_range_new_inaddr(netaddr, min_port, max_port);
+}
+
+/**
+ * This constructor creates a new ZSockAddrInetRange instance with the
+ * specified arguments.
+ *
+ * @param[in] addr address
+ * @param[in] min_port minimal port to bind to
+ * @param[in] max_port maximum port to bind to
+ *
+ * @returns the new ZSockAddr instance
+ **/
+ZSockAddr *
+z_sockaddr_inet_range_new_inaddr(struct in_addr addr, guint16 min_port, guint16 max_port)
+{
+  ZSockAddrInetRange *self;
+  
   self = g_new0(ZSockAddrInetRange, 1);
   self->refcnt = 1;
   self->flags = 0;
   self->salen = sizeof(struct sockaddr_in);
   self->sin.sin_family = AF_INET;
-  self->sin.sin_addr = netaddr;
+  self->sin.sin_addr = addr;
   self->sin.sin_port = 0;
   self->sa_funcs = &inet_range_sockaddr_funcs;
   if (max_port > min_port)
     {
-      self->last_port = (rand() % (max_port - min_port)) + min_port;
+      self->last_port = (guint16)(rand() % (max_port - min_port)) + min_port;
     }
   else if (max_port == min_port)
     {
@@ -588,13 +609,164 @@ z_sockaddr_inet_range_new(const gchar *ip, guint16 min_port, guint16 max_port)
   return (ZSockAddr *) self;
 }
 
+#ifndef G_OS_WIN32
+
+/**
+ * ZSockAddrInet6 is an implementation of the ZSockAddr interface,
+ * encapsulating an IPv6 host address and port number.
+ **/
+typedef struct _ZSockAddrInet6
+{
+  gint refcnt;
+  guint32 flags;
+  ZSockAddrFuncs *sa_funcs;
+  int salen;
+  struct sockaddr_in6 sin6;
+} ZSockAddrInet6;
+
+/**
+ * This function is the clonse callback for ZSockAddrInet sockets.
+ *
+ * @param[in] addr address to clone
+ * @param[in] wildcard whether the port information is important
+ *
+ * It copies the sockaddr contents and nulls port if wildcard is TRUE.
+ *
+ * @returns the cloned address
+ **/
+static ZSockAddr *
+z_sockaddr_inet6_clone(ZSockAddr *addr, gboolean wildcard)
+{
+  ZSockAddrInet6 *self = g_new0(ZSockAddrInet6, 1);
+  
+  memcpy(self, addr, sizeof(*self));
+  self->refcnt = 1;
+  if (wildcard)
+    self->sin6.sin6_port = 0;
+  
+  return (ZSockAddr *) self;
+}
+
+
+/**
+ * Format an IPv6 address into human readable form.
+ *
+ * @param[in]  addr the address
+ * @param[out] text buffer to receive the result
+ * @param[in]  n length of buffer
+ * 
+ * @returns a pointer to text
+ **/
+static gchar *
+z_sockaddr_inet6_format(ZSockAddr *addr, gchar *text, gulong n)
+{
+  ZSockAddrInet6 *self = (ZSockAddrInet6 *) addr;
+  char buf[64];
+
+  inet_ntop(AF_INET6, &self->sin6.sin6_addr, buf, sizeof(buf));
+  g_snprintf(text, n, "AF_INET6(%s:%d)",
+             buf,
+             htons(self->sin6.sin6_port));
+  return text;
+}
+
+/**
+ * Frees a ZSockAddrInet6 (or ZSockAddr) instance.
+ *
+ * @param[in] addr ZSockAddr instance
+ **/
+static void
+z_sockaddr_inet6_free(ZSockAddr *addr)
+{
+  g_free(addr);
+}
+
+/**
+ * Checks whether two IPv6 addresses of are equal.
+ *
+ * @param[in] addr the first address
+ * @param[in] o the second address
+ *
+ * @returns TRUE if both the IP address and port are equal.
+ **/
+static gboolean
+z_sockaddr_inet6_equal(ZSockAddr *addr, ZSockAddr *o)
+{
+  ZSockAddrInet6 *self = (ZSockAddrInet6 *) addr;
+  ZSockAddrInet6 *other = (ZSockAddrInet6 *) o;
+  
+  g_assert(self->sin6.sin6_family == AF_INET6);
+  g_assert(other->sin6.sin6_family == AF_INET6);
+  return memcmp(&self->sin6.sin6_addr, &other->sin6.sin6_addr, sizeof(self->sin6.sin6_addr)) == 0 && self->sin6.sin6_port == other->sin6.sin6_port;
+}
+
+/**
+ * ZSockAddrInet6 virtual methods.
+ **/
+static ZSockAddrFuncs inet6_sockaddr_funcs =
+{
+  z_sockaddr_inet_bind_prepare,
+  NULL,
+  z_sockaddr_inet6_clone,
+  z_sockaddr_inet6_format,
+  z_sockaddr_inet6_free,
+  z_sockaddr_inet6_equal
+};
+
+/**
+ * Allocate and initialize an IPv6 socket address.
+ *
+ * @param[in] ip text representation of an IP address
+ * @param[in] port port number in host byte order
+ *
+ * @returns ZSockAddrInet6 instance
+ **/
+ZSockAddr *
+z_sockaddr_inet6_new(gchar *ip, guint16 port)
+{
+  ZSockAddrInet6 *addr = g_new0(ZSockAddrInet6, 1);
+
+  addr->refcnt = 1;
+  addr->flags = 0;
+  addr->salen = sizeof(struct sockaddr_in6);
+  addr->sin6.sin6_family = AF_INET6;
+  inet_pton(AF_INET6, ip, &addr->sin6.sin6_addr);
+  addr->sin6.sin6_port = htons(port);
+  addr->sa_funcs = &inet6_sockaddr_funcs;
+
+  return (ZSockAddr *) addr;
+}
+
+/**
+ * Allocate and initialize an IPv6 socket address using libc sockaddr *
+ * structure.
+ *
+ * @param[in] sin6 the sockaddr_in6 structure to convert
+ *
+ * @returns ZSockAddrInet6 instance
+ **/
+ZSockAddr *
+z_sockaddr_inet6_new2(struct sockaddr_in6 *sin6)
+{
+  ZSockAddrInet6 *addr = g_new0(ZSockAddrInet6, 1);
+
+  addr->refcnt = 1;
+  addr->flags = 0;
+  addr->salen = sizeof(struct sockaddr_in6);
+  addr->sin6 = *sin6;
+  addr->sa_funcs = &inet6_sockaddr_funcs;
+
+  return (ZSockAddr *) addr;
+}
+#endif
+
 #ifndef G_OS_WIN32
 /* AF_UNIX socket address */
 
-/*
+/**
  * The ZSockAddrUnix class is an implementation of the ZSockAddr
  * interface encapsulating AF_UNIX domain socket names.
- */
+ **/
 typedef struct _ZSockAddrUnix
 {
   gint refcnt;
@@ -608,14 +780,14 @@ static GIOStatus z_sockaddr_unix_bind_prepare(int sock, ZSockAddr *addr, guint32
 static gchar *z_sockaddr_unix_format(ZSockAddr *addr, gchar *text, gulong n);
 
 /**
- * z_sockaddr_unix_clone: 
- * @addr: ZSockAddr instance to clone
- * @wildcard_clone: specifies whether the clone can be a wildcard
+ * This function is the clone callback for ZSockAddrUnix instances.
  *
- * This function is the clone callback for ZSockAddrUnix instances. It copies
- * the address information and returns a newly constructed, independent copy.
+ * @param[in] addr ZSockAddr instance to clone
+ * @param     wildcard_clone specifies whether the clone can be a wildcard (unused)
+ *
+ * It copies the address information and returns a newly constructed, independent copy.
  * 
- * Returns: the cloned instance
+ * @returns the cloned instance
  **/
 static ZSockAddr *
 z_sockaddr_unix_clone(ZSockAddr *addr, gboolean wildcard_clone G_GNUC_UNUSED)
@@ -628,22 +800,61 @@ z_sockaddr_unix_clone(ZSockAddr *addr, gboolean wildcard_clone G_GNUC_UNUSED)
   return (ZSockAddr *) self;
 }
 
+/**
+ * Checks whether two UNIX addresses of are equal.
+ *
+ * @param[in] addr the first address
+ * @param[in] o the second address
+ *
+ * @returns TRUE if the addresses are equal.
+ **/
+static gboolean
+z_sockaddr_unix_equal(ZSockAddr *addr, ZSockAddr *o)
+{
+  ZSockAddrUnix *self = (ZSockAddrUnix *) addr;
+  ZSockAddrUnix *other = (ZSockAddrUnix *) o;
+  
+  g_assert(self->saun.sun_family == AF_UNIX);
+  g_assert(other->saun.sun_family == AF_UNIX);
+  return strncmp(self->saun.sun_path, other->saun.sun_path, sizeof(self->saun.sun_path)) == 0;
+}
+
+/**
+ * ZSockAddrUnix virtual methods.
+ **/
 static ZSockAddrFuncs unix_sockaddr_funcs = 
 {
   z_sockaddr_unix_bind_prepare,
   NULL,
   z_sockaddr_unix_clone,
   z_sockaddr_unix_format,
-  NULL
+  NULL,
+  z_sockaddr_unix_equal
 };
 
 /* anonymous if name == NULL */
 
 /**
- * z_sockaddr_unix_new:
- * @name: filename
- * 
+ * This function checks whether the given ZSockAddr instance is in fact a
+ * ZSockAddrUnix instance.
+ *
+ * @param[in] s ZSockAddr instance
+ *
+ * @returns TRUE if s is a ZSockAddrUnix
+ **/
+
+gboolean
+z_sockaddr_unix_check(ZSockAddr *s)
+{
+  return s->sa_funcs == &unix_sockaddr_funcs;
+}
+
+/**
  * Allocate and initialize a ZSockAddrUnix instance.
+ *
+ * @param[in] name filename
+ *
+ * @returns a ZSockAddrUnix instance
  **/
 ZSockAddr *
 z_sockaddr_unix_new(const gchar *name)
@@ -668,12 +879,13 @@ z_sockaddr_unix_new(const gchar *name)
 }
 
 /**
- * z_sockaddr_unix_new2:
- * @saun: sockaddr_un structure to convert
- * @sunlen: size of saun
- *
  * Allocate and initialize a ZSockAddrUnix instance, using libc
- * sockaddr_un strucutre.
+ * sockaddr_un structure.
+ *
+ * @param[in] saun sockaddr_un structure to convert
+ * @param[in] sunlen size of saun
+ *
+ * @returns a ZSockAddrUnix instance
  **/
 ZSockAddr *
 z_sockaddr_unix_new2(struct sockaddr_un *saun, int sunlen)
@@ -683,22 +895,31 @@ z_sockaddr_unix_new2(struct sockaddr_un *saun, int sunlen)
   self->refcnt = 1;
   self->flags = 0;
   self->sa_funcs = &unix_sockaddr_funcs;
-  self->salen = sunlen;
-  self->saun = *saun;
+  memset(&self->saun, 0, sizeof(self->saun));
+  if (sunlen)
+    {
+      memcpy(&self->saun, saun, sunlen);
+    }
+  else
+    {
+      self->saun.sun_family = AF_UNIX;
+    }
+  self->salen = sizeof(struct sockaddr_un);
   return (ZSockAddr *) self;
 }
 
 
 /**
- * z_sockaddr_unix_bind_prepare:
- * @sock: fd of the socket to prepare for bind (not used)
- * @addr: address where we are binding to
- * @sock_flags: socket flags (ZSF_*)
- *
  * This function is used as a bind_prepare callback for ZSockAddrUnix
- * addresses. It currently unlinks the socket if it exists and is a socket.
+ * addresses. 
  *
- * Returns: GIOStatus to indicate success/failure
+ * @param     sock fd of the socket to prepare for bind (not used)
+ * @param[in] addr address where we are binding to
+ * @param     sock_flags socket flags (ZSF_*) (unused)
+ *
+ * It currently unlinks the socket if it exists and is a socket.
+ *
+ * @returns GIOStatus to indicate success/failure
  **/
 static GIOStatus
 z_sockaddr_unix_bind_prepare(int sock G_GNUC_UNUSED, ZSockAddr *addr, guint32 sock_flags G_GNUC_UNUSED)
@@ -718,16 +939,15 @@ z_sockaddr_unix_bind_prepare(int sock G_GNUC_UNUSED, ZSockAddr *addr, guint32 so
 }
 
 /**
- * z_sockaddr_unix_format:
- * @addr: ZSockAddrUnix instance
- * @text: buffer to format this address into
- * @n: size of text
- *
- * This function is the format callback for ZSockAddrUNix socket addresses
+ * This function is the format callback for ZSockAddrUnix socket addresses
  * and returns the human readable representation of the unix domain socket
  * address.
  *
- * Returns: the start of the buffer
+ * @param[in]  addr ZSockAddrUnix instance
+ * @param[out] text buffer to format this address into
+ * @param[in]  n size of text
+ *
+ * @returns the start of the buffer
  **/
 gchar *
 z_sockaddr_unix_format(ZSockAddr *addr, gchar *text, gulong n)
