@@ -14,7 +14,10 @@
  *
  ***************************************************************************/
 
-/* interface between Zorp and the SSL library */
+/**
+ * @file
+ * interface between Zorp and the SSL library
+ **/
 
 #include <zorp/ssl.h>
 #include <zorp/log.h>
@@ -39,13 +42,27 @@ static int ssl_initialized = 0;
 static GStaticMutex *ssl_mutexes;
 static int mutexnum;
 
+/**
+ * Fetch OpenSSL error code and generate a string interpretation of it.
+ *
+ * @param[out] buf buffer to put string into
+ * @param[in]  buflen size of buffer
+ *
+ * @returns buf
+ **/
 gchar *
 z_ssl_get_error_str(gchar *buf, int buflen)
 {
   const char *ls, *fs, *rs;
   unsigned long e, l, f, r;
+  unsigned long new_error = 0;
+  gint count = -1;
 
-  e = ERR_get_error();
+  do {
+    e = new_error;
+    new_error= ERR_get_error();
+    ++count; 
+  } while (new_error);
 
   l = ERR_GET_LIB(e);
   f = ERR_GET_FUNC(e);
@@ -55,16 +72,26 @@ z_ssl_get_error_str(gchar *buf, int buflen)
   fs = ERR_func_error_string(e);
   rs = ERR_reason_error_string(e);
 
-  g_snprintf(buf, buflen, "error:%08lX:%s:lib(%lu):%s:func(%lu):%s:reason(%lu)", e, ls ? ls : "(null)", l, fs ? fs : "(null)", f, rs ? rs : "(null)", r);
+  if (count)
+    g_snprintf(buf, buflen, "error:%08lX:%s:lib(%lu):%s:func(%lu):%s:reason(%lu), supressed %d messages", e, ls ? ls : "(null)", l, fs ? fs : "(null)", f, rs ? rs : "(null)", r, count);
+  else
+    g_snprintf(buf, buflen, "error:%08lX:%s:lib(%lu):%s:func(%lu):%s:reason(%lu)", e, ls ? ls : "(null)", l, fs ? fs : "(null)", f, rs ? rs : "(null)", r);
   return buf;
 }
 
 
+/**
+ * Callback used by OpenSSL to lock/unlock mutexes.
+ *
+ * @param[in] mode whether to lock or unlock the mutex
+ * @param[in] n number of mutex
+ * @param     file unused
+ * @param     line unused
+ **/
 static void
 z_ssl_locking_callback(int mode, int n, const char *file G_GNUC_UNUSED, int line G_GNUC_UNUSED)
 {
   z_enter();
-  
   if (n >= mutexnum)
     {
       /*LOG
@@ -73,6 +100,7 @@ z_ssl_locking_callback(int mode, int n, const char *file G_GNUC_UNUSED, int line
        */
       z_log(NULL, CORE_ERROR, 4, "SSL requested an out of bounds mutex; max='%d', n='%d'", mutexnum, n);
     }
+
   if (mode & CRYPTO_LOCK)
     {
       z_trace(NULL, "Mutex %d locked", n);
@@ -83,9 +111,12 @@ z_ssl_locking_callback(int mode, int n, const char *file G_GNUC_UNUSED, int line
       z_trace(NULL,  "Mutex %d unlocked", n);
       g_static_mutex_unlock(&ssl_mutexes[n]);
     }
-  z_leave();
+  z_return();
 }
 
+/**
+ * Initialize mutexes and set mutex locking callback for OpenSSL.
+ **/
 static void
 z_ssl_init_mutexes(void)
 {
@@ -93,33 +124,54 @@ z_ssl_init_mutexes(void)
   mutexnum = CRYPTO_num_locks();
   ssl_mutexes = g_new0(GStaticMutex, mutexnum);
   
+  z_enter();
   CRYPTO_set_locking_callback(z_ssl_locking_callback);
-  z_leave();
+  z_return();
 }
 
+/**
+ * Free OpenSSL error queue.
+ *
+ * @param thread unused
+ * @param user_data unused
+ *
+ * This function frees the OpenSSL error queue for the current thread.
+ **/
 static void
 z_ssl_remove_error_state(ZThread *thread G_GNUC_UNUSED, gpointer user_data G_GNUC_UNUSED)
 {
   ERR_remove_state(0);
 }
 
+/**
+ * Process ID callback function for OpenSSL.
+ *
+ * @returns pid
+ *
+ * @note See crypto_set_id_callback(3) for notes on what this is supposed to do.
+ **/
+static unsigned long
+z_ssl_get_id(void)
+{
+  return (unsigned long) g_thread_self();
+}
+
+/**
+ * Initialize OpenSSL library.
+ **/
 void
 z_ssl_init()
 {
   z_enter();
   if (ssl_initialized)
-    {
-      z_leave();
-      return;
-    }
+    z_return();
+  CRYPTO_set_id_callback(z_ssl_get_id);
   SSL_library_init();
   SSL_load_error_strings();
   SSLeay_add_all_algorithms();
 
 #if ZORPLIB_ENABLE_SSL_ENGINE
-
   ENGINE_load_builtin_engines();
-
   if (crypto_engine)
     {
       ENGINE *e;
@@ -143,7 +195,7 @@ z_ssl_init()
         }
       if (e)
         {
-          if (!ENGINE_set_default(e, ENGINE_METHOD_ALL)) 
+          if (!ENGINE_set_default(e, ENGINE_METHOD_ALL))
             {
               /*LOG
                 This message indicates that an error occurred during the crypto engine initialization.
@@ -167,15 +219,28 @@ z_ssl_init()
   z_ssl_init_mutexes();
   z_thread_register_stop_callback((GFunc) z_ssl_remove_error_state, NULL);
   ssl_initialized = 1;
-  z_leave();
+  z_return();
 }
 
+/**
+ * Deinitialize OpenSSL library.
+ **/
 void
 z_ssl_destroy(void)
 {
   ssl_initialized = 0;
 }
 
+/**
+ * Lookup name in X.509 certificate verification store.
+ *
+ * @param store certificate verification store
+ * @param type type
+ * @param name name
+ * @param obj object
+ *
+ * @todo FIXME-DOC: Document this once proper OpenSSL documentation is discovered.
+ **/
 static int
 z_ssl_x509_store_lookup(X509_STORE *store, int type,
                         X509_NAME *name, X509_OBJECT *obj)
@@ -187,8 +252,7 @@ z_ssl_x509_store_lookup(X509_STORE *store, int type,
   X509_STORE_CTX_init(&store_ctx, store, NULL, NULL);
   rc = X509_STORE_get_by_subject(&store_ctx, type, name, obj);
   X509_STORE_CTX_cleanup(&store_ctx);
-  z_leave();
-  return rc;
+  z_return(rc);
 }
 
 int
@@ -258,8 +322,7 @@ z_ssl_verify_crl(int ok,
           X509_STORE_CTX_set_error(ctx, X509_V_ERR_CRL_SIGNATURE_FAILURE);
           X509_OBJECT_free_contents(&obj);
           EVP_PKEY_free(pkey);
-          z_leave();
-          return FALSE;
+          z_return(FALSE);
         }
       EVP_PKEY_free(pkey);
 
@@ -274,8 +337,7 @@ z_ssl_verify_crl(int ok,
           
           X509_STORE_CTX_set_error(ctx, X509_V_ERR_ERROR_IN_CRL_NEXT_UPDATE_FIELD);
           X509_OBJECT_free_contents(&obj);
-          z_leave();
-          return FALSE;
+          z_return(FALSE);
         }
       if (i < 0)
         {
@@ -286,16 +348,13 @@ z_ssl_verify_crl(int ok,
           z_log(session_id, CORE_ERROR, 1, "CRL is expired; issuer='%s'", subject_name);
           X509_STORE_CTX_set_error(ctx, X509_V_ERR_CRL_HAS_EXPIRED);
           X509_OBJECT_free_contents(&obj);
-          z_leave();
-          return FALSE;
+          z_return(FALSE);
         }
       X509_OBJECT_free_contents(&obj);
     }
 
   memset((char *)&obj, 0, sizeof(obj));
-  
   rc = z_ssl_x509_store_lookup(crl_store, X509_LU_CRL, issuer, &obj);
-  
   crl = obj.data.crl;
   if (rc > 0 && crl != NULL)
     {
@@ -317,16 +376,13 @@ z_ssl_verify_crl(int ok,
                */
               z_log(session_id, CORE_ERROR, 1, "Certificate revoked by CRL; issuer='%s', serial=0x%lX",
                     issuer_name, serial);
-
               X509_OBJECT_free_contents(&obj);
-              z_leave();
-              return FALSE;
+              z_return(FALSE);
             }
         }
       X509_OBJECT_free_contents(&obj);
     }
-  z_leave();
-  return ok;
+  z_return(ok);
 }
                                                   
 int
@@ -341,10 +397,9 @@ z_ssl_verify_callback(int ok, X509_STORE_CTX *ctx)
   int forced_ok = FALSE;
 
   z_enter();
-  
   ssl = (SSL *) X509_STORE_CTX_get_app_data(ctx);
   verify_data = (ZSSLSession *) SSL_get_app_data(ssl);
-  
+
   xs = X509_STORE_CTX_get_current_cert(ctx);
   errnum = X509_STORE_CTX_get_error(ctx);
   errdepth = X509_STORE_CTX_get_error_depth(ctx);
@@ -378,7 +433,7 @@ z_ssl_verify_callback(int ok, X509_STORE_CTX *ctx)
       ok = TRUE;
       forced_ok = TRUE;
     }
-    
+
   if (ok && verify_data->crl_store)
     {
       ok = z_ssl_verify_crl(ok, xs, ctx, verify_data->crl_store, verify_data->session_id);
@@ -394,32 +449,26 @@ z_ssl_verify_callback(int ok, X509_STORE_CTX *ctx)
                 subject_name, issuer_name);
         }
     }
-  
-  if (ok && verify_data->verify_depth != -1)
+
+  if (ok && (verify_data->verify_depth != -1) &&
+      (errdepth > verify_data->verify_depth))
     {
-      if (errdepth > verify_data->verify_depth)
-        {
-          /*LOG
-            Certificate was verified successfully, but the length of the
-            verification chain is over configured limits.
-           */
-          z_log(verify_data->session_id, CORE_ERROR, 1, "Certificate chain is too long; subject='%s', issuer='%s' depth='%d' maxdepth='%d'", subject_name, issuer_name, errdepth, verify_data->verify_depth);
-          errnum = X509_V_ERR_CERT_CHAIN_TOO_LONG;
-          ok = FALSE;
-        }
+      /*LOG
+        Certificate was verified successfully, but the length of the
+        verification chain is over configured limits.
+       */
+      z_log(verify_data->session_id, CORE_ERROR, 1, "Certificate chain is too long; subject='%s', issuer='%s' depth='%d' maxdepth='%d'", subject_name, issuer_name, errdepth, verify_data->verify_depth);
+      errnum = X509_V_ERR_CERT_CHAIN_TOO_LONG;
+      ok = FALSE;
     }
 
   if (!ok)
-    {
-      /*LOG
-        This message indicates that certificate could not be verified, and
-	the certificate is treated as invalid.
-       */
-      z_log(verify_data->session_id, CORE_ERROR, 1, "Certificate verification error; subject='%s', issuer='%s', errcode='%d', error='%s'", subject_name, issuer_name, errnum, X509_verify_cert_error_string(errnum));
-    }
-
-  z_leave();
-  return ok || forced_ok;
+    /*LOG
+      This message indicates that certificate could not be verified, and
+      the certificate is treated as invalid.
+     */
+    z_log(verify_data->session_id, CORE_ERROR, 1, "Certificate verification error; subject='%s', issuer='%s', errcode='%d', error='%s'", subject_name, issuer_name, errnum, X509_verify_cert_error_string(errnum));
+  z_return(ok || forced_ok);
 }
 
 X509_STORE *
@@ -430,22 +479,17 @@ z_ssl_crl_store_create(gchar *crl_path)
 
   z_enter();        
   if ((store = X509_STORE_new()) == NULL)
-    {
-      z_leave();
-      return NULL;
-    }
+    z_return(NULL);
   if (crl_path != NULL)
     {
       if ((lookup = X509_STORE_add_lookup(store, X509_LOOKUP_hash_dir())) == NULL)
         {
           X509_STORE_free(store);
-          z_leave();
-          return NULL;
+          z_return(NULL);
         }
       X509_LOOKUP_add_dir(lookup, crl_path, X509_FILETYPE_PEM);
     }
-  z_leave();
-  return store;
+  z_return(store);
 }
 
 typedef struct _ZSSLCADirectory
@@ -467,7 +511,6 @@ z_ssl_dup_sk_x509_name(STACK_OF(X509_NAME) *old)
   int i;
 
   z_enter();
-
   sk = sk_X509_NAME_new_null();
   for (i = 0; i < sk_X509_NAME_num(old); i++)
     {
@@ -476,9 +519,7 @@ z_ssl_dup_sk_x509_name(STACK_OF(X509_NAME) *old)
       xn = sk_X509_NAME_value(old, i);
       sk_X509_NAME_push(sk, X509_NAME_dup(xn));
     }
-
-  z_leave();
-  return sk;
+  z_return(sk);
 }
 
 gboolean
@@ -490,13 +531,10 @@ z_ssl_set_trusted_ca_list(SSL_CTX *ctx, gchar *ca_path)
   STACK_OF(X509_NAME) *ca_file = NULL;
   const gchar *direntname;
   struct stat ca_stat;
-  
   GDir *dir;
    
   z_enter();
-  
   g_static_mutex_lock(&lock);
-
   if (ca_dir_hash == NULL)
     {
       ca_dir_hash = g_hash_table_new(g_str_hash, g_str_equal);
@@ -514,8 +552,7 @@ z_ssl_set_trusted_ca_list(SSL_CTX *ctx, gchar *ca_path)
             {
               SSL_CTX_set_client_CA_list(ctx, z_ssl_dup_sk_x509_name(ca_dir->contents));
               g_static_mutex_unlock(&lock);
-              z_leave();
-              return TRUE;
+              z_return(TRUE);
             }
           g_hash_table_remove(ca_dir_hash, orig_key);
           g_free(orig_key);
@@ -527,8 +564,7 @@ z_ssl_set_trusted_ca_list(SSL_CTX *ctx, gchar *ca_path)
   if (stat(ca_path, &ca_stat) < 0)
     {
       g_static_mutex_unlock(&lock);
-      z_leave();
-      return FALSE;
+      z_return(FALSE);
     }
   ca_dir = g_new0(ZSSLCADirectory, 1);
   ca_dir->modtime = ca_stat.st_mtime;
@@ -568,9 +604,7 @@ z_ssl_set_trusted_ca_list(SSL_CTX *ctx, gchar *ca_path)
   SSL_CTX_set_client_CA_list(ctx, z_ssl_dup_sk_x509_name(ca_dir->contents));
   g_dir_close(dir);
   g_static_mutex_unlock(&lock);
-  
-  z_leave();
-  return TRUE;
+  z_return(TRUE);
 }
 
 
@@ -578,12 +612,296 @@ z_ssl_set_trusted_ca_list(SSL_CTX *ctx, gchar *ca_path)
 #ifndef G_OS_WIN32
 
 static int
-z_ssl_password(char *buf, int size, int rwflag, void *userdata)
+z_ssl_password(char *buf G_GNUC_UNUSED, int size G_GNUC_UNUSED, int rwflag G_GNUC_UNUSED, void *userdata G_GNUC_UNUSED)
 {
   z_log(NULL, CORE_ERROR, 1, "Password protected key file detected;");
   return -1;
 }
 
+static SSL_CTX *
+z_ssl_create_ctx(char *session_id, int mode)
+{
+  SSL_CTX *ctx;
+  char buf[128];
+
+  z_enter();
+  if (mode == Z_SSL_MODE_CLIENT)
+    ctx = SSL_CTX_new(SSLv23_client_method());
+  else
+    ctx = SSL_CTX_new(SSLv23_server_method());
+
+  if (!ctx)
+    {
+      /*LOG
+        This message indicates that an SSL_CTX couldn't be allocated, it
+        usually means that Zorp has not enough memory. You might want
+        to check your ulimit settings and/or the available physical/virtual
+        RAM in your firewall host.
+       */
+      z_log(session_id, CORE_ERROR, 3, "Error allocating new SSL_CTX; error='%s'", z_ssl_get_error_str(buf, sizeof(buf)));
+      z_return(NULL);
+    }
+  SSL_CTX_set_options(ctx, SSL_OP_ALL);  
+  z_return(ctx);
+}
+
+static gboolean
+z_ssl_load_privkey_and_cert(char *session_id, SSL_CTX *ctx, gchar *key_file, gchar *cert_file)
+{
+  char buf[128];
+
+  z_enter();
+  if (key_file && key_file[0])
+    {
+      SSL_CTX_set_default_passwd_cb(ctx, z_ssl_password);
+      if (!SSL_CTX_use_PrivateKey_file(ctx, key_file, SSL_FILETYPE_PEM))
+        {
+          /*LOG
+            This message indicates that the given private key could not be 
+            loaded. Either the private key is not in PEM format, or 
+            the key data was encrypted.
+           */
+          z_log(session_id, CORE_ERROR, 3, "Error loading private key; keyfile='%s', error='%s'", key_file, z_ssl_get_error_str(buf, sizeof(buf)));
+          z_return(FALSE);
+        }
+      if (!SSL_CTX_use_certificate_file(ctx, cert_file, SSL_FILETYPE_PEM))
+        {
+          /*LOG
+            This message indicates that the given certificate file could not
+            be loaded. It might not be in PEM format or otherwise corrupted.
+           */
+          z_log(session_id, CORE_ERROR, 3, "Error loading certificate file; keyfile='%s', certfile='%s', error='%s'", key_file, cert_file, z_ssl_get_error_str(buf, sizeof(buf)));
+          z_return(FALSE);        
+        }
+      if (!SSL_CTX_check_private_key(ctx))
+        {
+          /*LOG
+            This message indicates that the private key and corresponding
+            certificate do not match.
+           */
+          z_log(session_id, CORE_ERROR, 3, "Certificate and private key mismatch; keyfile='%s', certfile='%s', error='%s'", key_file, cert_file, z_ssl_get_error_str(buf, sizeof(buf)));
+          z_return(FALSE);
+        }
+      /*LOG
+        This message reports that the given private key- and
+        certificate files were successfully loaded.
+       */
+      z_log(session_id, CORE_DEBUG, 6, "Certificate file successfully loaded; keyfile='%s', certfile='%s'", key_file, cert_file);
+    }
+  z_return(TRUE);
+}
+
+static gboolean
+z_ssl_set_privkey_and_cert(char *session_id, SSL_CTX *ctx, GString *key_pem, GString *cert_pem)
+{
+  char buf[128];
+
+  z_enter();
+  if (key_pem && key_pem->len)
+    {
+      EVP_PKEY *epk;
+      RSA *rsa;
+      BIO *bio;
+      X509 *x509;
+
+      bio = BIO_new_mem_buf(key_pem->str, key_pem->len);
+      if (!bio)
+        {
+          z_log(session_id, CORE_ERROR, 3, "Cannot create BIO for private key;");
+          z_return(FALSE);
+        }
+
+      rsa = PEM_read_bio_RSAPrivateKey(bio, NULL, NULL, NULL);
+      BIO_free(bio);
+      if (!rsa)
+        {
+          z_log(session_id, CORE_ERROR, 3, "Cannot parse rsa private key;");
+          z_return(FALSE);
+        }
+
+      epk = EVP_PKEY_new();
+      EVP_PKEY_assign_RSA(epk, rsa);
+      SSL_CTX_set_default_passwd_cb(ctx, z_ssl_password);
+      if (!SSL_CTX_use_PrivateKey(ctx, epk))
+        {
+          /*LOG
+            This message indicates that the given private key could not be 
+            loaded. Either the private key is not in PEM format, or 
+            the key data was encrypted.
+            */
+          z_log(session_id, CORE_ERROR, 3, "Error loading private key; error='%s'", z_ssl_get_error_str(buf, sizeof(buf)));
+          EVP_PKEY_free(epk);
+          z_return(FALSE);
+        }
+      EVP_PKEY_free(epk);
+
+      bio = BIO_new_mem_buf(cert_pem->str, cert_pem->len);
+      if (!bio)
+        {
+          z_log(session_id, CORE_ERROR, 3, "Cannot create BIO for certificate;");
+          z_return(FALSE);
+        }
+
+      x509 = PEM_read_bio_X509(bio, NULL, NULL, NULL);
+      BIO_free(bio);
+      if (!SSL_CTX_use_certificate(ctx, x509))
+        {
+          /*LOG
+            This message indicates that the given certificate file could not
+            be loaded. It might not be in PEM format or otherwise corrupted.
+            */
+          z_log(session_id, CORE_ERROR, 3, "Error loading certificate; error='%s'", z_ssl_get_error_str(buf, sizeof(buf)));
+          X509_free(x509);
+          z_return(FALSE);        
+        }
+      X509_free(x509);
+
+      if (!SSL_CTX_check_private_key(ctx))
+        {
+          /*LOG
+            This message indicates that the private key and corresponding
+            certificate do not match.
+            */
+          z_log(session_id, CORE_ERROR, 3, "Certificate and private key mismatch; error='%s'", z_ssl_get_error_str(buf, sizeof(buf)));
+          z_return(FALSE);
+        }
+      /*LOG
+        This message reports that the given private key- and
+        certificate files were successfully loaded.
+        */
+      z_log(session_id, CORE_DEBUG, 6, "Certificate successfully loaded;");
+    }
+  z_return(TRUE);
+}
+
+static gboolean
+z_ssl_load_ca_list(char *session_id, SSL_CTX *ctx, int mode, gchar *ca_dir, gchar *crl_dir, X509_STORE **crl_store)
+{
+  z_enter();
+  if (ca_dir && ca_dir[0])
+    {
+      if (mode == Z_SSL_MODE_SERVER)
+        {
+          if (!z_ssl_set_trusted_ca_list(ctx, ca_dir))
+            {
+              /*LOG
+                This message indicates that loading the trustable CA certificates
+                failed. The TLS peer is told which CAs we trust to ease the
+                selection of certificates. This could mean that one or more
+                of your CA certificates are invalid.
+               */
+              z_log(session_id, CORE_ERROR, 3, "Error loading trusted CA list; cadir='%s'", ca_dir);
+              z_return(FALSE);
+            }
+        }
+      if (access(ca_dir, R_OK | X_OK) < 0)
+        {
+          z_log(session_id, CORE_ERROR, 3, "Insufficient permissions to CA directory; cadir='%s', error='%s'", ca_dir, g_strerror(errno));
+          z_return(FALSE);
+        }
+      if (!SSL_CTX_load_verify_locations(ctx, NULL, ca_dir))
+        {
+          /*LOG
+            This message indicates that setting the trusted CA directory for
+            verification failed.
+           */
+          z_log(session_id, CORE_ERROR, 3, "Cannot add trusted CA directory as verify location; cadir='%s'", ca_dir);
+          z_return(FALSE);
+        }
+      if (crl_dir && crl_dir[0])
+        {
+          if (access(crl_dir, R_OK | X_OK) < 0)
+            {
+              z_log(session_id, CORE_ERROR, 3, "Insufficient permissions to CRL directory; crldir='%s', error='%s'", crl_dir, g_strerror(errno));
+              z_return(FALSE); 
+            }
+          *crl_store = z_ssl_crl_store_create(crl_dir);
+        }
+      else
+        {
+          /*LOG
+            This message reports that CRLs are not in use for certificate
+            verification.
+           */
+          z_log(session_id, CORE_DEBUG, 6, "Certificate Revocation Lists not available;");
+        }
+    }
+  z_return(TRUE);
+}
+
+static ZSSLSession *
+z_ssl_session_new_from_context(char *session_id, SSL_CTX *ctx, int verify_depth, int verify_type, X509_STORE *crl_store)
+{
+  ZSSLSession *self = NULL;
+  SSL *session;
+  int verify_mode = 0;
+
+  z_enter();
+  session = SSL_new(ctx);
+  if (!session)
+    {
+      /*LOG
+        This message indicates that creating the SSL session structure failed.
+        It's usually caused by out-of-memory conditions.
+       */
+      z_log(session_id, CORE_ERROR, 3, "Error creating SSL session from SSL_CTX;");
+      if (crl_store)
+        X509_STORE_free(crl_store);
+      z_return(NULL);
+    }
+  
+  self = g_new0(ZSSLSession, 1);
+  self->ref_cnt = 1;
+  self->ssl = session;  
+  self->session_id = session_id;
+  self->verify_type = verify_type;
+  self->verify_depth = verify_depth;
+  self->crl_store = crl_store;
+  SSL_set_app_data(session, self);
+
+  if (verify_type == Z_SSL_VERIFY_OPTIONAL || 
+      verify_type == Z_SSL_VERIFY_REQUIRED_UNTRUSTED)
+    verify_mode = SSL_VERIFY_PEER;
+  
+  if (verify_type == Z_SSL_VERIFY_REQUIRED_TRUSTED)
+    verify_mode = SSL_VERIFY_PEER | SSL_VERIFY_FAIL_IF_NO_PEER_CERT;
+  
+  if (verify_mode != 0)
+    SSL_set_verify(session, verify_mode, z_ssl_verify_callback);
+  
+  z_return(self);
+}
+
+ZSSLSession *
+z_ssl_session_new_inline(char *session_id, 
+                         int mode,
+                         GString *key_pem, 
+                         GString *cert_pem, 
+                         gchar *ca_dir, 
+                         gchar *crl_dir, 
+                         int verify_depth,
+                         int verify_type)
+{
+  ZSSLSession *self;
+  SSL_CTX *ctx;
+  X509_STORE *crl_store = NULL;
+  
+  z_enter();
+  ctx = z_ssl_create_ctx(session_id, mode);
+  if (!ctx)
+    z_return(NULL);
+
+  if (!z_ssl_set_privkey_and_cert(session_id, ctx, key_pem, cert_pem) ||
+      !z_ssl_load_ca_list(session_id, ctx, mode, ca_dir, crl_dir, &crl_store))
+    {
+      SSL_CTX_free(ctx);
+      z_return(NULL);
+    }
+  self = z_ssl_session_new_from_context(session_id, ctx, verify_depth, verify_type, crl_store);
+  SSL_CTX_free(ctx);
+  z_return(self);
+}
+
 ZSSLSession *
 z_ssl_session_new(char *session_id, 
                   int mode,
@@ -596,174 +914,22 @@ z_ssl_session_new(char *session_id,
 {
   ZSSLSession *self;
   SSL_CTX *ctx;
-  SSL *session;
   X509_STORE *crl_store = NULL;
-  int verify_mode = 0;
-  int ret;
-  char buf[128];
   
   z_enter();
-  
-  if (mode == Z_SSL_MODE_CLIENT)
-    ctx = SSL_CTX_new(SSLv23_client_method());
-  else
-    ctx = SSL_CTX_new(SSLv23_server_method());
-
+  ctx = z_ssl_create_ctx(session_id, mode);
   if (!ctx)
-    {
-      /*LOG
-        This message indicates that an SSL_CTX couldn't be allocated, it
-        usually means that Zorp has not enough memory. You might want
-        to check your ulimit settings and/or the available physical/virtual
-        RAM in your firewall host.
-       */
-      z_log(session_id, CORE_ERROR, 3, "Error allocating new SSL_CTX; error='%s'", z_ssl_get_error_str(buf, sizeof(buf)));
-      z_leave();
-      return NULL;
-    }
-  SSL_CTX_set_options(ctx, SSL_OP_ALL);  
+    z_return(NULL);
 
-  if (key_file && key_file[0])
+  if (!z_ssl_load_privkey_and_cert(session_id, ctx, key_file, cert_file) ||
+      !z_ssl_load_ca_list(session_id, ctx, mode, ca_dir, crl_dir, &crl_store))
     {
-      SSL_CTX_set_default_passwd_cb(ctx, z_ssl_password);
-      ret = SSL_CTX_use_PrivateKey_file(ctx, key_file, SSL_FILETYPE_PEM);
-      if (!ret)
-        {
-          /*LOG
-            This message indicates that the given private key could not be 
-            loaded. Either the private key is not in PEM format, or 
-            the key data was encrypted.
-           */
-          z_log(session_id, CORE_ERROR, 3, "Error loading private key; keyfile='%s', error='%s'", key_file, z_ssl_get_error_str(buf, sizeof(buf)));
-          SSL_CTX_free(ctx);
-          z_leave();
-          return NULL;
-        }
-      ret = SSL_CTX_use_certificate_file(ctx, cert_file, SSL_FILETYPE_PEM);
-      if (!ret)
-        {
-          /*LOG
-            This message indicates that the given certificate file could not
-            be loaded. It might not be in PEM format or otherwise corrupted.
-           */
-          z_log(session_id, CORE_ERROR, 3, "Error loading certificate file; keyfile='%s', certfile='%s', error='%s'", key_file, cert_file, z_ssl_get_error_str(buf, sizeof(buf)));
-          SSL_CTX_free(ctx);
-          z_leave();
-          return NULL;        
-        }
-      ret = SSL_CTX_check_private_key(ctx);
-      if (!ret)
-        {
-          /*LOG
-            This message indicates that the private key and corresponding
-            certificate do not match.
-           */
-          z_log(session_id, CORE_ERROR, 3, "Certificate and private key mismatch; keyfile='%s', certfile='%s', error='%s'", key_file, cert_file, z_ssl_get_error_str(buf, sizeof(buf)));
-          SSL_CTX_free(ctx);
-          z_leave();
-          return NULL;
-        }
-      /*LOG
-        This message reports that the given private key- and
-        certificate files were successfully loaded.
-       */
-      z_log(session_id, CORE_DEBUG, 6, "Certificate file successfully loaded; keyfile='%s', certfile='%s'", key_file, cert_file);
+      SSL_CTX_free(ctx);
+      z_return(NULL);
     }
-  if (ca_dir && ca_dir[0])
-    {
-      if (mode == Z_SSL_MODE_SERVER)
-        {
-          if (!z_ssl_set_trusted_ca_list(ctx, ca_dir))
-            {
-              /*LOG
-                This message indicates that loading the trustable CA certificates
-                failed. The TLS peer is told which CAs we trust to ease the
-                selection of certificates. This could mean that one or more
-                of your CA certificates are invalid.
-               */
-              z_log(session_id, CORE_ERROR, 3, "Error loading trusted CA list; cadir='%s'", ca_dir);
-              SSL_CTX_free(ctx);
-              z_leave();
-              return NULL;
-            }
-        }
-      if (access(ca_dir, R_OK | X_OK) < 0)
-        {
-          z_log(session_id, CORE_ERROR, 3, "Insufficient permissions to CA directory; cadir='%s', error='%s'", ca_dir, g_strerror(errno));
-          SSL_CTX_free(ctx);
-          z_leave();
-          return NULL;
-        }
-      if (!SSL_CTX_load_verify_locations(ctx, NULL, ca_dir))
-        {
-          /*LOG
-            This message indicates that setting the trusted CA directory for
-            verification failed.
-           */
-          z_log(session_id, CORE_ERROR, 3, "Cannot add trusted CA directory as verify location; cadir='%s'", ca_dir);
-          SSL_CTX_free(ctx);
-          z_leave();
-          return NULL;
-        }
-      if (crl_dir && crl_dir[0])
-        {
-          if (access(crl_dir, R_OK | X_OK) < 0)
-            {
-              z_log(session_id, CORE_ERROR, 3, "Insufficient permissions to CRL directory; crldir='%s', error='%s'", crl_dir, g_strerror(errno));
-            }
-          else
-            {
-              crl_store = z_ssl_crl_store_create(crl_dir);
-            }
-        }
-      else
-        {
-          /*LOG
-            This message reports that CRLs are not in use for certificate
-            verification.
-           */
-          z_log(session_id, CORE_DEBUG, 6, "Certificate Revocation Lists not available;");
-        }
-    }
-  session = SSL_new(ctx);
+  self = z_ssl_session_new_from_context(session_id, ctx, verify_depth, verify_type, crl_store);
   SSL_CTX_free(ctx);
-  if (!session)
-    {
-      /*LOG
-        This message indicates that creating the SSL session structure failed.
-        It's usually caused by out-of-memory conditions.
-       */
-      z_log(session_id, CORE_ERROR, 3, "Error creating SSL session from SSL_CTX;");
-      if (crl_store)
-        X509_STORE_free(crl_store);
-      z_leave();
-      return NULL;
-    }
-  
-  self = g_new0(ZSSLSession, 1);
-  self->ref_cnt = 1;
-  self->ssl = session;  
-  self->session_id = session_id;
-  self->verify_type = verify_type;
-  self->verify_depth = verify_depth;
-  self->crl_store = crl_store;
-  SSL_set_app_data(session, self);
-
-  if (verify_type == Z_SSL_VERIFY_OPTIONAL || 
-      verify_type == Z_SSL_VERIFY_REQUIRED_UNTRUSTED)
-    {
-      verify_mode = SSL_VERIFY_PEER;
-    }
-  if (verify_type == Z_SSL_VERIFY_REQUIRED_TRUSTED)
-    {
-      verify_mode = SSL_VERIFY_PEER | SSL_VERIFY_FAIL_IF_NO_PEER_CERT;
-    }
-  
-  if (verify_mode != 0)
-    SSL_set_verify(session, verify_mode, z_ssl_verify_callback);
-  
-  z_leave();
-  return self;
+  z_return(self);
 }
 #else
 
@@ -778,12 +944,10 @@ z_ssl_session_new(char *session_id,
   ZSSLSession *self;
   SSL_CTX *ctx;
   SSL *session;
-  X509_STORE *crl_store = NULL;
   int verify_mode = 0;
   char buf[128];
   
   z_enter();
-  
   if (mode == Z_SSL_MODE_CLIENT)
     ctx = SSL_CTX_new(SSLv23_client_method());
   else
@@ -798,17 +962,13 @@ z_ssl_session_new(char *session_id,
         RAM in your firewall host.
        */
       z_log(session_id, CORE_ERROR, 3, "Error allocating new SSL_CTX; error='%s'", z_ssl_get_error_str(buf, sizeof(buf)));
-      z_leave();
-      return NULL;
+      z_return(NULL);
     }
   SSL_CTX_set_options(ctx, SSL_OP_ALL);  
-/*-----------------------------------------------------------------------------*/  
-  crl_store = store;
   
-  if(store)
-    SSL_CTX_set_cert_store(ctx, crl_store);
+  if (store)
+    SSL_CTX_set_cert_store(ctx, store);
   
-/*-----------------------------------------------------------------------------*/  
   session = SSL_new(ctx);
   SSL_CTX_free(ctx);
   if (!session)
@@ -818,43 +978,33 @@ z_ssl_session_new(char *session_id,
         It's usually caused by out-of-memory conditions.
        */
       z_log(session_id, CORE_ERROR, 3, "Error creating SSL session from SSL_CTX;");
-      if (crl_store)
-        X509_STORE_free(crl_store);
-      z_leave();
-      return NULL;
+      if (store)
+        X509_STORE_free(store);
+      z_return(NULL);
     }
-  
+
+  /* FIXME: CRL handling */  
   self = g_new0(ZSSLSession, 1);
   self->ref_cnt = 1;
   self->ssl = session;  
   self->session_id = session_id;
   self->verify_type = verify_type;
   self->verify_depth = verify_depth;
-  self->crl_store = crl_store;
   SSL_set_app_data(session, self);
 
   if (verify_type == Z_SSL_VERIFY_OPTIONAL || 
       verify_type == Z_SSL_VERIFY_REQUIRED_UNTRUSTED)
-    {
-      verify_mode = SSL_VERIFY_PEER;
-    }
+    verify_mode = SSL_VERIFY_PEER;
   if (verify_type == Z_SSL_VERIFY_REQUIRED_TRUSTED)
-    {
-      verify_mode = SSL_VERIFY_PEER | SSL_VERIFY_FAIL_IF_NO_PEER_CERT;
-    }
-  
+    verify_mode = SSL_VERIFY_PEER | SSL_VERIFY_FAIL_IF_NO_PEER_CERT;
+
   if (verify_mode != 0)
     SSL_set_verify(session, verify_mode, z_ssl_verify_callback);
-  
-  z_leave();
-  return self;
+  z_return(self);
 
 }
 #endif
 
-
-
-
 ZSSLSession *
 z_ssl_session_new_ssl(SSL *ssl)
 {
@@ -871,12 +1021,12 @@ static void
 z_ssl_session_free(ZSSLSession *self)
 {
   z_enter();
-
   /* free verify_data */
   SSL_free(self->ssl);
+  if (self->crl_store)
+    X509_STORE_free(self->crl_store);
   g_free(self);
-  
-  z_leave();
+  z_return();
 }
 
 ZSSLSession *
@@ -889,7 +1039,7 @@ z_ssl_session_ref(ZSSLSession *self)
 void
 z_ssl_session_unref(ZSSLSession *self)
 {
-  if (--self->ref_cnt == 0)
+  if (self && --self->ref_cnt == 0)
     {
       z_ssl_session_free(self);
     }
@@ -912,7 +1062,6 @@ z_stream_bio_write(BIO *bio, const char *buf, int buflen)
   gsize write_size;
 
   z_enter();
-
   if (buf != NULL)
     {
       ret = z_stream_write(self->stream, buf, buflen, &write_size, NULL);
@@ -921,18 +1070,12 @@ z_stream_bio_write(BIO *bio, const char *buf, int buflen)
       if (ret == G_IO_STATUS_AGAIN)
         {
           BIO_set_retry_write(bio);
-        
-	  z_leave();
-	  return -1;
-        }
-      else if (ret != G_IO_STATUS_NORMAL)
-        {
-      	  z_leave();
-	  return -1;
+	  z_return(-1);
         }
+      if (ret != G_IO_STATUS_NORMAL)
+        z_return(-1);
     }
-  z_leave();
-  return rc;
+  z_return(rc);
 }
 
 int
@@ -944,7 +1087,6 @@ z_stream_bio_read(BIO *bio, char *buf, int buflen)
   gsize read_size;
 
   z_enter();
-
   if (buf != NULL)
     {
       ret = z_stream_read(self->stream, buf, buflen, &read_size, NULL);
@@ -953,24 +1095,14 @@ z_stream_bio_read(BIO *bio, char *buf, int buflen)
       if (ret == G_IO_STATUS_AGAIN)
         {
           BIO_set_retry_read(bio);
-        
-	  z_leave();
-	  return -1;
-        }
-      else if (ret == G_IO_STATUS_EOF)
-        {
-          z_leave();
-          return 0;
-        }
-      else if (ret != G_IO_STATUS_NORMAL)
-        {
-          z_leave();
-          return -1;
+	  z_return(-1);
         }
+      if (ret == G_IO_STATUS_EOF)
+        z_return(0);
+      if (ret != G_IO_STATUS_NORMAL)
+        z_return(-1);
     }
-
-  z_leave();
-  return rc;
+  z_return(rc);
 }
 
 int
@@ -979,12 +1111,9 @@ z_stream_bio_puts(BIO *bio, const char *str)
   int n, ret;
 
   z_enter();
-        
   n = strlen(str);
   ret = z_stream_bio_write(bio, str, n);
-
-  z_leave();
-  return(ret);
+  z_return(ret);
 }
 
 long
@@ -993,19 +1122,21 @@ z_stream_bio_ctrl(BIO *bio, int cmd, long num, void *ptr G_GNUC_UNUSED)
   long ret = 1;
 
   z_enter();
-
   switch (cmd)
     {
     case BIO_CTRL_GET_CLOSE:
       ret = bio->shutdown;
       break;
+      
     case BIO_CTRL_SET_CLOSE:
       bio->shutdown = (int)num;
       break;
+      
     case BIO_CTRL_DUP:
     case BIO_CTRL_FLUSH:
       ret = 1;
       break;
+      
     case BIO_CTRL_RESET:
     case BIO_C_FILE_SEEK:
     case BIO_C_FILE_TELL:
@@ -1018,23 +1149,18 @@ z_stream_bio_ctrl(BIO *bio, int cmd, long num, void *ptr G_GNUC_UNUSED)
       ret = 0;
       break;
     }
-    
-  z_leave();
-  return(ret);
+  z_return(ret);
 }
 
 int
 z_stream_bio_create(BIO *bio)
 {
   z_enter();
-
   bio->init = 1;
   bio->num = 0;
   bio->ptr = NULL;
   bio->flags = 0;
-  
-  z_leave();
-  return(1);
+  z_return(1);
 }
 
 int
@@ -1043,22 +1169,15 @@ z_stream_bio_destroy(BIO *bio)
   ZStreamBio *self = (ZStreamBio *)bio;
 
   z_enter();
-
   if (self == NULL)
-    {
-      z_leave();
-      return(0);
-    }
-
+    z_return(0);
   if (self->super.shutdown)
     {
       z_stream_shutdown(self->stream, 2, NULL);
       bio->init = 0;
       bio->flags = 0;
     }
-    
-  z_leave();    
-  return(1);
+  z_return(1);
 }
 
 BIO_METHOD z_ssl_bio_method = 
@@ -1081,11 +1200,8 @@ z_ssl_bio_new(ZStream *stream)
   ZStreamBio *self = g_new0(ZStreamBio, 1);
 
   z_enter();
-  
   self->super.method = &z_ssl_bio_method;
   self->stream = stream;
   self->super.init = 1;
-  
-  z_leave();
-  return (BIO *)self;
+  z_return((BIO *)self);
 }
