@@ -26,17 +26,25 @@
 
 #ifdef G_OS_WIN32
 #  include <winsock2.h>
+#  include <zorp/io.h>
 #else
+#  include <fcntl.h>
 #  include <sys/socket.h>
 #  include <sys/poll.h>
 #endif
 
+/**
+ * ZStream-derived class that can be used to read/write an fd.
+ *
+ * On Windows, this can only be a WinSock fd.
+ **/
 typedef struct _ZStreamFD
 {
   ZStream super;
 
   GIOChannel *channel;
   gint fd;
+  gint keepalive;
   GPollFD pollfd;
 #ifdef G_OS_WIN32
   int winsock_event;
@@ -46,6 +54,9 @@ typedef struct _ZStreamFD
 
 extern ZClass ZStreamFD__class;
 
+/**
+ * ZStreamFD extra context data.
+ **/
 typedef struct _ZStreamFDExtra
 {
   gboolean nonblock;
@@ -65,14 +76,9 @@ z_stream_fd_watch_prepare(ZStream *s, GSource *src G_GNUC_UNUSED, gint *timeout)
 #endif
 
   z_enter();
-
   *timeout = -1;
-  
   if (mypollfd->revents)
-    {
-      z_leave();
-      return TRUE;
-    }
+    z_return(TRUE);
 
 #ifdef G_OS_WIN32
   if (mystream->super.want_write)
@@ -113,10 +119,8 @@ z_stream_fd_watch_prepare(ZStream *s, GSource *src G_GNUC_UNUSED, gint *timeout)
 	to the Balabit QA Team (devel@balabit.com).
        */
       z_log(NULL, CORE_ERROR, 1, "Failed to set up WinSock event for z_stream_fd; error_code='%d'", z_errno_get());
-      z_leave();
-      return FALSE;
+      z_return(FALSE);
     }
-  
   mypollfd->events = G_IO_IN;
 
 #else
@@ -133,9 +137,7 @@ z_stream_fd_watch_prepare(ZStream *s, GSource *src G_GNUC_UNUSED, gint *timeout)
     mypollfd->events |= G_IO_PRI;
   
 #endif
-
-  z_leave();
-  return FALSE;
+  z_return(FALSE);
 }
 
 static gboolean
@@ -184,15 +186,12 @@ z_stream_fd_watch_check(ZStream *s, GSource *src G_GNUC_UNUSED)
 #endif
 
   poll_condition = mypollfd->revents;
-  
-  z_leave();
-  return poll_condition != 0;
+  z_return(poll_condition != 0);
 }
 
 static gboolean
 z_stream_fd_watch_dispatch(ZStream *s, GSource *src)
 {
-  ZStreamSource *self = (ZStreamSource *) src;
   ZStreamFD *mystream = (ZStreamFD *) s;
   GPollFD *mypollfd = &mystream->pollfd;
   GIOCondition poll_cond = mypollfd->revents;
@@ -205,13 +204,13 @@ z_stream_fd_watch_dispatch(ZStream *s, GSource *src)
 
   mypollfd->revents = 0;
   
-  if (poll_cond & (G_IO_ERR | G_IO_HUP))
+  if (poll_cond & (G_IO_ERR | G_IO_HUP) && rc)
     {
       if (mystream->super.want_read)
-        rc = (*mystream->super.read_cb)(self->stream, poll_cond, mystream->super.user_data_read);
+        rc = (*mystream->super.read_cb)(&mystream->super, poll_cond, mystream->super.user_data_read);
         
       else if (mystream->super.want_write)
-        rc = (*mystream->super.write_cb)(self->stream, poll_cond, mystream->super.user_data_write);
+        rc = (*mystream->super.write_cb)(&mystream->super, poll_cond, mystream->super.user_data_write);
         
       else
         {
@@ -222,18 +221,17 @@ z_stream_fd_watch_dispatch(ZStream *s, GSource *src)
             of interoperability problem between your operating system
             and Zorp.
            */
-          z_log(mystream->super.name, CORE_ERROR, 4, "POLLERR or POLLHUP was received on an inactive fd; fd='%d'", mypollfd->fd);
-          g_source_destroy(&self->super);
+          z_log(mystream->super.name, CORE_ERROR, 4, "Internal error, POLLERR or POLLHUP was received on an inactive fd; fd='%d'", mypollfd->fd);
+          g_source_destroy(src);
         }
-      z_leave();
-      return rc;
+      z_return(rc);
     }
     
-  if (mystream->super.want_read && (poll_cond & G_IO_IN))
+  if (mystream->super.want_read && (poll_cond & G_IO_IN) && rc)
     {
       if (mystream->super.read_cb)
         {
-          rc = (*mystream->super.read_cb)(self->stream, poll_cond, mystream->super.user_data_read);
+          rc = (*mystream->super.read_cb)(&mystream->super, poll_cond, mystream->super.user_data_read);
         }
       else
         {
@@ -249,7 +247,7 @@ z_stream_fd_watch_dispatch(ZStream *s, GSource *src)
     {
       if (mystream->super.write_cb)
         {
-          rc &= (*mystream->super.write_cb)(self->stream, poll_cond, mystream->super.user_data_write);
+          rc &= (*mystream->super.write_cb)(&mystream->super, poll_cond, mystream->super.user_data_write);
         }
       else
         {
@@ -265,7 +263,7 @@ z_stream_fd_watch_dispatch(ZStream *s, GSource *src)
     {
       if (mystream->super.pri_cb)
         {
-          rc &= (*mystream->super.pri_cb)(self->stream, poll_cond, mystream->super.user_data_pri);
+          rc &= (*mystream->super.pri_cb)(&mystream->super, poll_cond, mystream->super.user_data_pri);
         }
       else
         {
@@ -276,16 +274,19 @@ z_stream_fd_watch_dispatch(ZStream *s, GSource *src)
           z_log(mystream->super.name, CORE_ERROR, 3, "Internal error, no pri callback is set;");
         }
     }
-
-  z_leave();
-  return rc;
+  z_return(rc);
 }
 
-void
-z_stream_fd_watch_finalize(ZStream *s G_GNUC_UNUSED, GSource *source G_GNUC_UNUSED)
-{
-}
 
+/**
+ * Wait for the fd encapsulated by self.
+ *
+ * @param[in] self ZStreamFD instance
+ * @param[in] cond events to wait for
+ * @param[in] timeout timeout
+ *
+ * On win32 select is used. On other platforms poll is used.
+ **/
 #ifndef G_OS_WIN32
 
 static gboolean
@@ -296,13 +297,9 @@ z_stream_wait_fd(ZStreamFD *self, guint cond, gint timeout)
   GIOFlags flags;
 
   z_enter();
-  
   flags = g_io_channel_get_flags(self->channel);
   if ((flags & G_IO_FLAG_NONBLOCK) || timeout == -2)
-    {
-      z_leave();
-      return TRUE;
-    }
+    z_return(TRUE);
   errno = 0;
   pfd.fd = self->fd;
   pfd.events = cond;
@@ -311,15 +308,11 @@ z_stream_wait_fd(ZStreamFD *self, guint cond, gint timeout)
     {
       res = poll(&pfd, 1, timeout);
       if (res == 1)
-        {
-          z_leave();
-          return TRUE;
-        }
+        z_return(TRUE);
     }
   while (res == -1 && errno == EINTR);
   errno = ETIMEDOUT;
-  z_leave();
-  return FALSE;
+  z_return(FALSE);
 }
 
 #else
@@ -333,14 +326,13 @@ z_stream_wait_fd(ZStreamFD *self, guint cond, gint timeout)
   GIOFlags flags;
 
   z_enter();
-
   if (cond & G_IO_IN)
     {
       u_long bytes;
   
       ioctlsocket(self->fd, FIONREAD, &bytes);
       if (bytes > 0)
-        return TRUE;
+        z_return(TRUE);
     }
 
   FD_ZERO(&rf);
@@ -353,10 +345,7 @@ z_stream_wait_fd(ZStreamFD *self, guint cond, gint timeout)
 
   flags = g_io_channel_get_flags(self->channel);
   if ((flags & G_IO_FLAG_NONBLOCK) || timeout == -2)
-    {
-      z_leave();
-      return TRUE;
-    }
+    z_return(TRUE);
 
   if (timeout > 0)
     {
@@ -371,73 +360,88 @@ z_stream_wait_fd(ZStreamFD *self, guint cond, gint timeout)
   if ((res == 0) || (res == SOCKET_ERROR))
     {
       z_errno_set(ETIMEDOUT);
-      z_leave();
-      return FALSE;
+      z_return(FALSE);
     }
-  z_leave();
-  return TRUE;
+  z_return(TRUE);
 }
 
 #endif
 
+/**
+ * Read from the fd encapsulated by a ZStreamFD instance.
+ *
+ * @param[in]  stream ZStreamFD instance
+ * @param[in]  buf buffer to read to
+ * @param[in]  count number of bytes to ask for
+ * @param[out] bytes_read number of bytes actually read will be put here
+ * @param[out] error error value
+ *
+ * @returns GIOStatus value
+ **/
 static GIOStatus
 z_stream_fd_read_method(ZStream *stream,
-		        gchar *buf,
+		        void *buf,
 		        gsize count,
 		        gsize *bytes_read,
 		        GError **error)
 {
   ZStreamFD *self = (ZStreamFD *) stream;
   GIOStatus res;
+  GError *local_error = NULL;
 
   z_enter();
+  
   g_return_val_if_fail ((error == NULL) || (*error == NULL), G_IO_STATUS_ERROR);
 
   if (!z_stream_wait_fd(self, G_IO_IN | G_IO_HUP, self->super.timeout))
     {
-      /*LOG
-        This message indicates that reading from the channel timed out.
-       */
-      z_log(self->super.name, CORE_ERROR, 1, "Channel read timed out; fd='%d'", self->fd);
-      g_set_error (error, G_IO_CHANNEL_ERROR,
-                          G_IO_CHANNEL_ERROR_FAILED,
-                          "Channel read timed out");
-      z_leave();
-      return G_IO_STATUS_ERROR;
+      g_set_error(&local_error, G_IO_CHANNEL_ERROR, G_IO_CHANNEL_ERROR_FAILED, "Channel read timed out");
+      res = G_IO_STATUS_ERROR;
     }
-
-  res = g_io_channel_read_chars(self->channel, buf, count, bytes_read, error);
-  if (res == G_IO_STATUS_ERROR)
+  else
     {
-      /*LOG
-        This message indicates that read() failed because of the given
-        reason.
-       */
-      z_log(self->super.name, CORE_ERROR, 1, "Channel read failed; error='%s'", g_strerror(errno));
-      z_leave();
-      return res;
+      res = g_io_channel_read_chars(self->channel, buf, count, bytes_read, &local_error);
     }
-
-  self->super.bytes_recvd += *bytes_read;
-
-  if (res != G_IO_STATUS_AGAIN)
+  if (!(self->super.umbrella_state & G_IO_IN))
     {
-      /*LOG
-        This message reports the number of bytes read from the given fd.
-       */
-      z_log(self->super.name, CORE_DUMP, 7, "Reading channel; fd='%d', count='%zd'", self->fd, *bytes_read);
+      /* low-level logging if we're not the toplevel stream */
       
-      if (z_log_enabled(CORE_DUMP, 9))
-        z_data_dump(self->super.name, buf, *bytes_read);
+      if (res == G_IO_STATUS_NORMAL)
+        {
+          /*LOG
+            This message reports the number of bytes read from the given fd.
+           */
+          z_log(self->super.name, CORE_DUMP, 8, "Reading channel; fd='%d', count='%zd'", self->fd, *bytes_read);
+          z_log_data_dump(self->super.name, CORE_DUMP, 10, buf, *bytes_read);
+        }
+      else if (res == G_IO_STATUS_EOF)
+        {
+          /*LOG
+            This message reports that EOF was read from the given fd.
+           */
+          z_log(self->super.name, CORE_DUMP, 8, "Reading EOF on channel; fd='%d'", self->fd);
+        }
     }
 
-  z_leave();
-  return res;
+  if (local_error)
+    g_propagate_error(error, local_error);
+  z_return(res);
 }
 
+/**
+ * Write to the fd encapsulated by a ZStreamFD instance.
+ *
+ * @param[in]  stream ZStreamFD instance
+ * @param[in]  buf buffer to write the contents of
+ * @param[in]  count number of bytes to write
+ * @param[out] bytes_written actual number of bytes written will be returned here
+ * @param[out] error error value
+ *
+ * @returns GIOStatus value
+ **/
 static GIOStatus
 z_stream_fd_write_method(ZStream *stream,
-		         const gchar *buf,
+		         const void *buf,
 		         gsize count,
 		         gsize *bytes_written,
 		         GError **error)
@@ -448,45 +452,40 @@ z_stream_fd_write_method(ZStream *stream,
   z_enter();
   g_return_val_if_fail ((error == NULL) || (*error == NULL), G_IO_STATUS_ERROR);
   
-  /*LOG
-    This message indicates that the given number of bytes is to be written
-    to the given fd.
-   */
-  z_log(self->super.name, CORE_DUMP, 7, "Writing channel; fd='%d', count='%zd'", self->fd, count);
-
   if (!z_stream_wait_fd(self, G_IO_OUT | G_IO_HUP, self->super.timeout))
     {
-      /*LOG
-        This message indicates that writing to the given fd timed out.
-       */
-      z_log(self->super.name, CORE_ERROR, 1, "Channel write timed out; fd='%d'", self->fd);
-      g_set_error (error, G_IO_CHANNEL_ERROR,
-                          G_IO_CHANNEL_ERROR_FAILED,
-                          "Channel write timed out");
-      z_leave();
-      return G_IO_STATUS_ERROR;
+      g_set_error(error, G_IO_CHANNEL_ERROR, G_IO_CHANNEL_ERROR_FAILED, "Channel write timed out");
+      z_return(G_IO_STATUS_ERROR);
     }
 
   res = g_io_channel_write_chars(self->channel, buf, count, bytes_written, error);
-  if (res == G_IO_STATUS_ERROR)
+  if (!(self->super.umbrella_state & G_IO_OUT))
     {
-      /*LOG
-        This message indicates that some I/O error occurred in
-        the write() system call.
-       */
-      z_log(self->super.name, CORE_ERROR, 1, "Channel write failed; error='%s'", g_strerror(errno));
-      z_leave();
-      return res;
+      /* low-level logging if we're not the toplevel stream */
+      
+      if (res != G_IO_STATUS_AGAIN)
+        {
+          /*LOG
+            This message reports the number of bytes read from the given fd.
+           */
+          z_log(self->super.name, CORE_DUMP, 8, "Writing channel; fd='%d', count='%zd'", self->fd, *bytes_written);
+          z_log_data_dump(self->super.name, CORE_DUMP, 10, buf, *bytes_written);
+        }
     }
-
-  if (z_log_enabled(CORE_DEBUG, 9))
-    z_data_dump(self->super.name, buf, *bytes_written);
-
-  self->super.bytes_sent += *bytes_written;
-  z_leave();
-  return res;
+  z_return(res);
 }
 
+/**
+ * Write priority data to the fd encapsulated by a ZStreamFD instance.
+ *
+ * @param[in]  stream ZStreamFD instance
+ * @param[in]  buf buffer to write the contents of
+ * @param[in]  count number of bytes to write
+ * @param[out] bytes_written actual number of bytes written will be returned here
+ * @param[out] error error value
+ *
+ * @returns GIOStatus value
+ **/
 GIOStatus
 z_stream_fd_write_pri_method(ZStream *stream,
 		             const gchar *buf,
@@ -499,32 +498,30 @@ z_stream_fd_write_pri_method(ZStream *stream,
 
   z_enter();
   g_return_val_if_fail ((error == NULL) || (*error == NULL), G_IO_STATUS_ERROR);  
-
   do
     {
       if (!z_stream_wait_fd(self, G_IO_OUT | G_IO_HUP, self->super.timeout))
-	{
-	  /*LOG
-	    This message indicates that send() timed out.
-	   */
-	  z_log(self->super.name, CORE_ERROR, 1, "Send timed out; fd='%d'", self->fd);
+        {
+          /*LOG
+            This message indicates that send() timed out.
+           */
+          z_log(self->super.name, CORE_ERROR, 1, "Send timed out; fd='%d'", self->fd);
           g_set_error (error, G_IO_CHANNEL_ERROR,
-                              G_IO_CHANNEL_ERROR_FAILED,
-                              "Channel send timed out");
-          z_leave();
-	  return G_IO_STATUS_ERROR;
-	}
+                       G_IO_CHANNEL_ERROR_FAILED,
+                       "Channel send timed out");
+          z_return(G_IO_STATUS_ERROR);
+        }
       res = send(self->fd, buf, count, MSG_OOB);
       if (res == -1)
         {
-	  if (!z_errno_is(EINTR) && !z_errno_is(EAGAIN))
-	    {
-	      /*LOG
-	        This message indicates that some I/O error occurred in the
-	        send() system call.
-	       */
-	      z_log(self->super.name, CORE_ERROR, 1, "Send failed; attempt='%d', error='%s'", attempt++, g_strerror(errno));
-	    }
+          if (!z_errno_is(EINTR) && !z_errno_is(EAGAIN))
+            {
+              /*LOG
+                This message indicates that some I/O error occurred in the
+                send() system call.
+               */
+              z_log(self->super.name, CORE_ERROR, 1, "Send failed; attempt='%d', error='%s'", attempt++, g_strerror(errno));
+            }
         }
     }
   while (res == -1 && z_errno_is(EINTR));
@@ -532,31 +529,36 @@ z_stream_fd_write_pri_method(ZStream *stream,
   if (res >= 0)
     {
       *bytes_written = res;
-
       self->super.bytes_sent += res;
-      z_leave();
-      return G_IO_STATUS_NORMAL;
+      z_return(G_IO_STATUS_NORMAL);
     }
-    
+
   if (z_errno_is(EAGAIN))
-    {
-      z_leave();
-      return G_IO_STATUS_AGAIN;
-    }
-  else
-    {
-      g_clear_error(error);
-      g_set_error (error, G_IO_CHANNEL_ERROR,
-                          g_io_channel_error_from_errno (z_errno_get()),
-                          strerror (z_errno_get()));
-      z_leave();
-      return G_IO_STATUS_ERROR;
-    }
-  
-  z_leave();
-  return G_IO_STATUS_ERROR;
+    z_return(G_IO_STATUS_AGAIN);
+
+  g_clear_error(error);
+  g_set_error (error, G_IO_CHANNEL_ERROR,
+               g_io_channel_error_from_errno (z_errno_get()),
+               strerror (z_errno_get()));
+  z_return(G_IO_STATUS_ERROR);
 }
 
+/**
+ * Close the socket associated with a ZStreamFD instance.
+ *
+ * @param[in]  stream ZStreamFD instance
+ * @param[in]  i HOW argument to shutdown
+ * @param[out] error error value
+ *
+ * shutdown(2) will be called on the fd.
+ *
+ * The action to perform is specified by i as follows:
+ * - i == 0: Stop receiving data.
+ * - i == 1: Stop trying to transmit data.
+ * - i == 2: Stop both reception and transmission.
+ *
+ * @returns GIOStatus value
+ **/
 static GIOStatus
 z_stream_fd_shutdown_method(ZStream *stream, int i, GError **error)
 {
@@ -594,44 +596,92 @@ z_stream_fd_shutdown_method(ZStream *stream, int i, GError **error)
       g_set_error (error, G_IO_CHANNEL_ERROR,
                           g_io_channel_error_from_errno (z_errno_get()),
                           strerror (z_errno_get()));
-      z_leave();
-      return G_IO_STATUS_ERROR;
+      z_return(G_IO_STATUS_ERROR);
     }
-  z_leave();
-  return G_IO_STATUS_NORMAL;
+  z_return(G_IO_STATUS_NORMAL);
 }
 
+/**
+ * Close the fd associated with a ZStreamFD instance.
+ *
+ * @param[in]  s ZStreamFD instance
+ * @param[out] error error value
+ *
+ * @returns GIOStatus value
+ **/
 static GIOStatus
-z_stream_fd_close_method(ZStream *stream, GError **error)
+z_stream_fd_close_method(ZStream *s, GError **error)
 {
-  ZStreamFD *self = (ZStreamFD *) stream;
+  ZStreamFD *self = Z_CAST(s, ZStreamFD);
+  GIOStatus res;
 
   z_enter();
-  /*LOG
-    This message indicates that the given fd is to be closed.
-   */
-
-  z_log(self->super.name, CORE_DEBUG, 6, "Closing channel; fd='%d'", self->fd);
-  z_leave();
-  return g_io_channel_shutdown(self->channel, TRUE, error);
+  res = Z_SUPER(self, ZStream)->close(s, error);
+  if (res == G_IO_STATUS_NORMAL)
+    res = g_io_channel_shutdown(self->channel, TRUE, error);
+  z_return(res);
 }
 
+/**
+ * Process stream control calls.
+ *
+ * @param[in]      s ZStream instance
+ * @param[in]      function function selector
+ * @param[in, out] value parameter to function
+ * @param[in]      vlen length of value
+ *
+ * @returns TRUE on success
+ **/
 static gboolean
 z_stream_fd_ctrl_method(ZStream *s, guint function, gpointer value, guint vlen)
 {
   ZStreamFD *self = Z_CAST(s, ZStreamFD);
   
   z_enter();
-  
   switch (ZST_CTRL_MSG(function))
     {
+    case ZST_CTRL_SET_CLOSEONEXEC:
+      if (vlen == sizeof(gboolean))
+        {
+          gboolean cloexec = *((gboolean *)value);
+          gint fd = self->fd;
+          gint res;
+
+#ifndef G_OS_WIN32
+          if (cloexec)
+            res = fcntl(fd, F_SETFD, FD_CLOEXEC);
+          else
+            res = fcntl(fd, F_SETFD, ~FD_CLOEXEC);
+#else
+          res = 0;
+#endif    
+          if (res >= 0)
+            z_return(TRUE);
+
+          /*LOG
+            This message indicates that an internal error occurred, during setting CLOSE_ON_EXEC mode
+            on a stream. Please report this event to the Balabit QA Team (devel@balabit.com).
+           */
+          z_log(NULL, CORE_ERROR, 4, "Internal error, during setting CLOSE_ON_EXEC mode;");
+        }
+      else
+        {
+          /*LOG
+            This message indicates that an internal error occurred, during setting CLOSE_ON_EXEC
+            mode on a stream, because the size of the parameter is wrong. Please report this
+            event to the Balabit QA Team (devel@balabit.com).
+           */
+          z_log(NULL, CORE_ERROR, 4, "Internal error, bad parameter is given for setting CLOSE_ON_EXEC mode; size='%d'", vlen);
+        }
+      break;
+
     case ZST_CTRL_SET_NONBLOCK:
       if (vlen == sizeof(gboolean))
         {
           gboolean nonblock = *((gboolean *)value);
           GIOStatus ret;
           GIOFlags flags;
-          
+
           flags = g_io_channel_get_flags(self->channel);
 #ifndef G_OS_WIN32
           if (nonblock)
@@ -639,153 +689,144 @@ z_stream_fd_ctrl_method(ZStream *s, guint function, gpointer value, guint vlen)
           else
             ret = g_io_channel_set_flags(self->channel, flags & ~G_IO_FLAG_NONBLOCK, NULL);
 #else
+          z_fd_set_nonblock(self->fd, nonblock);
           ret = G_IO_STATUS_NORMAL;
 #endif          
           if (ret == G_IO_STATUS_NORMAL)
-            {
-              z_leave();
-              return TRUE;
-            }
-          else
-	    /*LOG
-	      This message indicates that an internal error, during setting NONBLOCK mode on
-	      a stream. Please report this event to the Balabit QA Team (devel@balabit.com).
-	     */
-            z_log(NULL, CORE_ERROR, 4, "Internal error, during setting NONBLOCK mode;");
+            z_return(TRUE);
+
+          /*LOG
+            This message indicates that an internal error, during setting NONBLOCK mode on
+            a stream. Please report this event to the Balabit QA Team (devel@balabit.com).
+           */
+          z_log(NULL, CORE_ERROR, 4, "Internal error, during setting NONBLOCK mode;");
         }
       else
-	/*LOG
-	  This message indicates that an internal error occurred, during setting NONBLOCK mode
-	  on a stream, because the size of the parameter is wrong. Please report this event to
-	  the Balabit QA Team (devel@balabit.com).
-	 */
+        /*LOG
+          This message indicates that an internal error occurred, during setting NONBLOCK mode
+          on a stream, because the size of the parameter is wrong. Please report this event to
+          the Balabit QA Team (devel@balabit.com).
+         */
         z_log(NULL, CORE_ERROR, 4, "Internal error, bad parameter is given for setting NONBLOCK mode; size='%d'", vlen);
-
       break;
+
     case ZST_CTRL_GET_NONBLOCK:
       if (vlen == sizeof(gboolean))
         {
           GIOFlags flags;
-          
+
           flags = g_io_channel_get_flags(self->channel);
           *((gboolean *) value) = !!(flags & G_IO_FLAG_NONBLOCK);
-          z_leave();
-          return TRUE;
+          z_return(TRUE);
         }
-      else
-	/*LOG
-	  This message indicates that an internal error occurred, during getting NONBLOCK mode status
-	  on a stream, because the size of the parameter is wrong. Please report this event to
-	  the Balabit QA Team (devel@balabit.com).
-	 */
-        z_log(NULL, CORE_ERROR, 4, "Internal error, bad parameter is given for getting the NONBLOCK mode; size='%d'", vlen);
+      /*LOG
+        This message indicates that an internal error occurred, during getting NONBLOCK mode status
+        on a stream, because the size of the parameter is wrong. Please report this event to
+        the Balabit QA Team (devel@balabit.com).
+       */
+      z_log(NULL, CORE_ERROR, 4, "Internal error, bad parameter is given for getting the NONBLOCK mode; size='%d'", vlen);
       break;
+
     case ZST_CTRL_GET_FD:
       if (vlen == sizeof(gint))
         {
           *((gint *)value) = self->fd;
+          z_return(TRUE);
+        }
+      /*LOG
+        This message indicates that an internal error occurred, during getting the FD of a stream,
+        because the size of the parameter is wrong. Please report this event to the Balabit QA
+        Team (devel@balabit.com).
+       */
+      z_log(NULL, CORE_ERROR, 4, "Internal error, bad parameter is given for getting the FD; size='%d'", vlen);
+      break;
+
+    case ZST_CTRL_GET_BROKEN:
+#ifndef G_OS_WIN32
+      if (vlen == sizeof(gboolean))
+        {
+          gchar buf[1];
+          gint res;
+
+          res = recv(self->fd, buf, sizeof(buf), MSG_PEEK | MSG_DONTWAIT);
+          if ((res < 0 && (z_errno_is(EAGAIN) || z_errno_is(ENOTSOCK))) || res > 0)
+            *((gboolean *)value) = FALSE;
+          else
+            *((gboolean *)value) = TRUE;
+          z_return(TRUE);
+        }
+      /*LOG
+        This message indicates that an internal error occurred, during getting the FD of a stream,
+        because the size of the parameter is wrong. Please report this event to the Balabit QA
+        Team (devel@balabit.com).
+       */
+      z_log(NULL, CORE_ERROR, 4, "Internal error, bad parameter is given for getting the broken state; size='%d'", vlen);
+#else
+      z_log(NULL, CORE_ERROR, 4, "Internal error, this feature is not supported on Win32;");
+      z_return(FALSE);
+#endif          
+      break;
+
+    case ZST_CTRL_GET_KEEPALIVE:
+      if (vlen == sizeof(gint))
+        {
+          *((gint *)value) = self->keepalive;
           z_leave();
           return TRUE;
         }
       else
-	/*LOG
-	  This message indicates that an internal error occurred, during getting the FD of a stream,
-	  because the size of the parameter is wrong. Please report this event to the Balabit QA
-	  Team (devel@balabit.com).
-	 */
-        z_log(NULL, CORE_ERROR, 4, "Internal error, bad parameter is given for getting the FD; size='%d'", vlen);
-        
+        /*LOG
+          This message indicates that an internal error occurred, during getting the FD of a stream,
+          because the size of the parameter is wrong. Please report this event to the Balabit QA
+          Team (devel@balabit.com).
+          */
+        z_log(NULL, CORE_ERROR, 4, "Internal error, bad parameter is given for getting the KEEPALIVE option; size='%d'", vlen);
       break;
-    default:
-      if (z_stream_ctrl_method(s, function, value, vlen))
+
+    case ZST_CTRL_SET_KEEPALIVE:
+      if (vlen == sizeof(gint))
         {
+          self->keepalive = *((gint *)value);
           z_leave();
           return TRUE;
         }
+      else
+        /*LOG
+          This message indicates that an internal error occurred, during getting the FD of a stream,
+          because the size of the parameter is wrong. Please report this event to the Balabit QA
+          Team (devel@balabit.com).
+          */
+        z_log(NULL, CORE_ERROR, 4, "Internal error, bad parameter is given for setting the KEEPALIVE option; size='%d'", vlen);
+      break;
+
+    default:
+      if (z_stream_ctrl_method(s, function, value, vlen))
+        z_return(TRUE);
       /*LOG
         This message indicates that an internal error occurred, because an invalid control was called.
-	Please report this event to the Balabit QA Team (devel@balabit.com).
+        Please report this event to the Balabit QA Team (devel@balabit.com).
        */
       z_log(NULL, CORE_ERROR, 4, "Internal error, unknown stream ctrl; ctrl='%d'", ZST_CTRL_MSG(function));
       break;
     }
-  z_leave();
-  return FALSE;
+  z_return(FALSE);
 }
 
-
-
 static void
-z_stream_fd_attach_source_method(ZStream *stream, GMainContext *context)
+z_stream_fd_attach_source_method(ZStream *s, GMainContext *context)
 {
-  ZStreamFD *self = (ZStreamFD *) stream;
+  ZStreamFD *self = (ZStreamFD *) s;
   z_enter();
-  if (!stream->source)
-    {
-      stream->source = z_stream_source_new(stream);
-      
+  
+  Z_SUPER(self, ZStream)->attach_source(s, context);
 #ifdef G_OS_WIN32
-      self->pollfd.fd = self->winsock_event;
-      z_trace(NULL, "WinSock: Event #%d created at %s line %d", self->pollfd.fd, __FILE__, __LINE__);
+  self->pollfd.fd = self->winsock_event;
+  z_trace(NULL, "WinSock: Event #%d created at %s line %d", self->pollfd.fd, __FILE__, __LINE__);
 #else
-      self->pollfd.fd = ((ZStreamFD *) stream)->fd;
+  self->pollfd.fd = self->fd;
 #endif
-
-      g_source_add_poll(stream->source, &self->pollfd);
-
-      g_source_attach(stream->source, context);
-    }
-  
-  z_leave();
-  return;
-}
-
-static void
-z_stream_fd_detach_source_method(ZStream *stream)
-{
-  GSource *source;
-  
-  z_enter();
-  if (stream->source)
-    {
-      source = stream->source;
-      stream->source = NULL;
-      g_source_destroy(source);
-      g_source_unref(source);
-    }
-    
-  z_leave();
-}
-
-/* destructor */
-static void
-z_stream_fd_free_method(ZObject *s)
-{
-  ZStreamFD *self = Z_CAST(s, ZStreamFD);
-  time_t time_close;
-
-  z_enter();
-  time_close = time(NULL);
-  
-  /*LOG
-    This message contains accounting information on the given channel. It
-    reports the number of seconds the fd was open and the number of
-    bytes sent/received on this channel.
-   */
-  z_log(self->super.name, CORE_ACCOUNTING, 4,
-        "accounting info; type='stream', duration='%.0f', sent='%" G_GUINT64_FORMAT "', received='%" G_GUINT64_FORMAT "'", 
-        difftime(time_close, self->super.time_open),
-        self->super.bytes_sent,
-        self->super.bytes_recvd);
-
-#ifdef G_OS_WIN32 
-  z_trace(NULL, "WinSock: Event #%d destroyed at %s line %d", self->winsock_event, __FILE__, __LINE__); 
-  WSACloseEvent(self->winsock_event); 
-#endif   
-
-  g_io_channel_unref(self->channel);
-  z_stream_free_method(s);
-  z_leave();
+  g_source_add_poll(s->source, &self->pollfd);
+  z_return();
 }
 
 static gsize
@@ -820,6 +861,57 @@ z_stream_fd_extra_restore_method(ZStream *s, gpointer extra)
   return ofs + sizeof(ZStreamFDExtra);
 }
 
+
+
+/**
+ * Allocate and initialize a ZStreamFD instance with the given fd and name.
+ *
+ * @param[in] fd fd to wrap this stream around
+ * @param[in] name name to identify to stream in logs
+ *
+ * @returns ZStream instance
+ **/
+ZStream *
+z_stream_fd_new(gint fd, gchar *name)
+{
+  ZStreamFD *self;
+
+  z_enter();
+  self = Z_CAST(z_stream_new(Z_CLASS(ZStreamFD), name, G_IO_IN|G_IO_OUT), ZStreamFD);
+  self->fd = fd;
+#ifdef G_OS_WIN32
+  self->winsock_event = (int)WSACreateEvent();
+  self->channel = g_io_channel_win32_new_socket(fd);
+#else
+  self->channel = g_io_channel_unix_new(fd);
+#endif
+  self->keepalive = 0;
+  g_io_channel_set_encoding(self->channel, NULL, NULL);
+  g_io_channel_set_buffered(self->channel, FALSE);
+  g_io_channel_set_close_on_unref(self->channel, FALSE);
+  z_return(&self->super);
+}
+
+/** destructor */
+static void
+z_stream_fd_free_method(ZObject *s)
+{
+  ZStreamFD *self = Z_CAST(s, ZStreamFD);
+
+  z_enter();
+#ifdef G_OS_WIN32 
+  z_trace(NULL, "WinSock: Event #%d destroyed at %s line %d", self->winsock_event, __FILE__, __LINE__); 
+  WSACloseEvent(self->winsock_event); 
+#endif   
+  g_io_channel_unref(self->channel);
+  z_stream_free_method(s);
+  z_return();
+}
+
+
+/**
+ * ZStreamFD virtual methods.
+ **/
 static ZStreamFuncs 
 z_stream_fd_funcs =
 {
@@ -836,17 +928,22 @@ z_stream_fd_funcs =
   z_stream_fd_ctrl_method,
 
   z_stream_fd_attach_source_method,
-  z_stream_fd_detach_source_method,
+  NULL, /* detach_source */
   z_stream_fd_watch_prepare,
   z_stream_fd_watch_check,
   z_stream_fd_watch_dispatch,
-  z_stream_fd_watch_finalize,
+  NULL, /* watch_finalize */
   
   z_stream_fd_extra_get_size_method,
   z_stream_fd_extra_save_method,
-  z_stream_fd_extra_restore_method
+  z_stream_fd_extra_restore_method,
+  NULL,
+  NULL
 };
 
+/**
+ * ZStreamFD class descriptor.
+ **/
 ZClass ZStreamFD__class =
 {
   Z_CLASS_HEADER,
@@ -855,38 +952,3 @@ ZClass ZStreamFD__class =
   sizeof(ZStreamFD),
   &z_stream_fd_funcs.super,
 };
-
-
-/*+
-
-  Allocate and initialize a ZStreamFD instance with the given fd and name.
-
-  Parameters:
-    fd        fd to wrap this stream around
-    name      name to identify to stream in logs
-  
-  Returns:
-    ZStream instance
-
-  +*/
-ZStream *
-z_stream_fd_new(gint fd, gchar *name)
-{
-  ZStreamFD *self;
-
-  z_enter();
-  self = Z_CAST(z_stream_new(Z_CLASS(ZStreamFD), name, NULL, Z_STREAM_FLAG_READ|Z_STREAM_FLAG_WRITE), ZStreamFD);
-  self->fd = fd;
-#ifdef G_OS_WIN32
-  self->winsock_event = (int)WSACreateEvent();
-  self->channel = g_io_channel_win32_new_socket(fd);
-#else
-  self->channel = g_io_channel_unix_new(fd);
-#endif
-  g_io_channel_set_encoding(self->channel, NULL, NULL);
-  g_io_channel_set_buffered(self->channel, FALSE);
-  g_io_channel_set_close_on_unref(self->channel, FALSE);
-
-  z_leave();
-  return &self->super;
-}
